% FGHC -> FLENG compiler (baseline version)

:- dynamic([declaration/1, no_warnings/0, include_path/1]).
:- dynamic([moded/2, included/1]).

:- if(current_prolog_flag(dialect, gprolog)).
% GNU
:- initialization(main).

main :-
    argument_list(Args),
    main(Args).
:- endif.

:- if(current_prolog_flag(dialect, swi)).
% SWI
format_to_atom(Out, Fstr, Args) :-
    format_to_codes(Fstr, Args, S),
    atom_codes(Out, S).

start :-
    set_prolog_flag(on_error, halt),
    current_prolog_flag(argv, Args),
    main(Args).

g_assign(Var, Val) :- nb_setval(Var, Val).
g_read(Var, Val) :- nb_getval(Var, Val).

save_to_file :- 
    set_prolog_flag(double_quotes, codes),
    qsave_program('fghc2fl.swi', [stand_alone(true), goal(start)]), 
    halt(0).

file_exists(File) :- exists_file(File).
:- endif.

:- if(current_prolog_flag(dialect, yap)).
% YAP
:- initialization(main).

:- use_module(library(lists)).
:- use_module(library(charsio)).

format_to_atom(Out, Fstr, Args) :-
    format_to_chars(Fstr, Args, S),
    atom_codes(Out, S).

save_to_file :- save('fghc2fl.yap'), halt(0).

start :- 
    set_prolog_flag(syntax_errors, error),
    set_prolog_flag(unknown, error),
    unix(argv(Args)), 
    catch(main(Args), E, quit(E)).

quit(E) :- 
    print_message(error, E),
    halt(1).

numbervars(Term) :- numbervars(Term, 0, _).

g_assign(Var, Val) :- nb_setval(Var, Val).
g_read(Var, Val) :- nb_getval(Var, Val).
:- endif.

:- include('version.pl').

main(['-v']) :-
    fleng_version(V),
    format('fghc2fl ~a\n', [V]),
    halt(0).
main(['-I', Dir|Args]) :-
    asserta(include_path(Dir)),
    main(Args).
main(['-i', Fname|Args]) :-
    assertz(included(Fname)),
    main(Args).
main(['-w'|Args]) :-
    assertz(no_warnings),
    main(Args).
main([Infile, Outfile]) :-
    op(200, fx, !),
    op(700, xfx, '\\=:='),
    op(500, yfx, '><'),
    op(400, yfx, '\\\\'),
    op(700, yfx, '@'),
    op(900, xfy, '&'),
    op(880, xfx, ':='),
    op(880, xfx, '<='),
    op(880, xfx, '+='),
    op(880, xfx, '-='),
    op(880, xfx, '*='),
    op(880, xfx, '/='),
    op(880, xfx, '<=='),
    op(880, xfx, '=>'),
    op(0, xfx, '^'),
    op(0, fx, initialization),
    set_prolog_flag(double_quotes, codes),
    assertz(include_path('.')),
    g_assign(counter, 1),
    see(Infile),
    findall(Name, included(Name), Incs),
    include_files(Incs, Prg),
    seen,
    collect_groups(Prg, Groups),
    translate(Groups, TG),
    open(Outfile, write, Out),
    fleng_version(V),
    format(Out, '%% GENERATED BY fghc2fl [~a] FROM ~a\n\n', [V, Infile]),
    write_declarations(Out),
    write_translation(TG, Out),
    format(Out, '%% END OF FILE\n', []),
    close(Out),
    halt(0).
main(_) :- 
    format(user_error, 'usage: fghc2fl [-v] [-w] [-I DIRECTORY] [-i FILENAME] INFILE OUTFILE\n', []),
    halt(1).

dbg(X) :- format('~q\n', [X]).
dbg(F, X) :- format(F, [X]).

error(Fstr, Args) :-
    format(user_error, Fstr, Args),
    halt(1).

warning(Fstr, Args) :-
    \+no_warnings,
    format(user_error, 'Warning: ', []),
    format(user_error, Fstr, Args).
warning(_, _).

read_program(PRG) :-
    read_term(TERM, [singletons(SL)]),
    report_singletons(TERM, SL, TERM),
    numbervars(TERM),
    read_program(TERM, PRG).

read_program(end_of_file, []).
read_program(-DECL, PRG) :-
    process_declaration(DECL, PRG).
read_program(TERM, [C|PRG]) :-
    canonicalize(TERM, NA, Args, Guards, Body),
    apply_modes(NA, Args, Guards, Body, C),
    read_program(PRG).

apply_modes(NA, Args1, Guards, Body1, c(NA, Args, Guards, Body)) :-
    moded(NA, Modes),
    !,
    extract_variables((Args1, Guards, Body1), Vars),
    apply_modes2(Args1, Modes, Vars, Args, Body1, Body).
apply_modes(NA, Args, Guards, Body, c(NA, Args, Guards, Body)).    

apply_modes2([], _, _, [], B, B).
apply_modes2([X|Args], ['^'|Modes], Vars, ['$VAR'(V)|Args2], B1, B) :-
    available_variable(Vars, V),
    moded_result_assign(V, X, Assign),
    apply_modes2(Args, Modes, [V|Vars], Args2, (B1, Assign), B).
apply_modes2([X|Args], ['?'|Modes], Vars, [X|Args2], B1, B) :-
    apply_modes2(Args, Modes, Vars, Args2, B1, B).

moded_result_assign(V, X, '$assign'(X, '$VAR'(V))) :- atomic(X).
moded_result_assign(V, X, '$VAR'(V) = X).

report_singletons(_, [], _).
report_singletons(-_, _, _).
report_singletons(_, SL, _) :-
    format(user_error, 'Warning: singleton variables ', []),
    memberchk(Name = _, SL),
    format(user_error, '~a ', [Name]),
    fail.
report_singletons(_, _, Term) :-
    format(user_error, 'in term:\n\n  ~q\n', [Term]).

% declarations

process_declaration(include(Fname), PRG) :-
    atom(Fname),
    process_declaration(include([Fname]), PRG).
process_declaration(include(Fnames), PRG) :-
    include_files(Fnames, PRG).
process_declaration(mode(Mode), PRG) :-
    process_mode(Mode),
    read_program(PRG).
process_declaration(Decl, PRG) :-
    assertz(declaration(Decl)),
    read_program(PRG).

include_files([], Prg) :- read_program(Prg).
include_files([Fname|Fnames], Prg) :-
    current_input(In),
    find_include_file(Fname, Fname2),
    see(Fname2),
    read_program(Included),
    seen,
    set_input(In),
    append(Included, Next, Prg),
    include_files(Fnames, Next).
include_files(X, _) :-
    error('invalid include file specification: ~w\n', [X]).

find_include_file(Fname1, Fname) :-
    include_path(Dir),
    check_file(Dir, Fname1, Fname),
    !.
find_include_file(Fname, _) :-
    error('included file does not exist: ~a\n', [Fname]).

check_file(Dir, Fname1, Fname) :-
    format_to_atom(Fname, '~a/~a', [Dir, Fname1]),     
    file_exists(Fname),
    !.
check_file(Dir, Fname1, Fname) :-
    format_to_atom(Fname, '~a/~a.ghc', [Dir, Fname1]),     
    file_exists(Fname).

check_machine_declaration(ring).
check_machine_declaration(torus).
check_machine_declaration(M) :-
    error('unsupported machine topology: ~w\n', [M]).

process_mode(T) :-
    T =.. [Name|Modes],
    length(Modes, A),
    assertz(moded(Name/A, Modes)).
process_mode(T) :-
    error('invalid mode declaration: ~w\n', [T]).

% translation

write_declarations(Out) :-
    declaration(Decl),
    format(Out, '-~q.\n', [Decl]),
    fail.
write_declarations(Out) :- nl(Out).

collect_groups([], []).
collect_groups([T|PRG], [[T|G]|GROUPS]) :-
    T = c(NA, _, _, _),
    filter_group(PRG, NA, NA, G, NG),
    collect_groups(NG, GROUPS).

filter_group([], _, _, [], []).
filter_group([T|PRG], NA, NA, [T|G], NG) :-
    T = c(NA, _, _, _),
    !,
    filter_group(PRG, NA, NA, G, NG).
filter_group([T|PRG], N/A, _, [T|G], NG) :-
    T = c(N/A, _, _, _),
    !,
    format(user_error, 'Warning: non-contiguous process definition for ~a/~d\n', [N, A]),
    filter_group(PRG, NA, NA, G, NG).
filter_group([T|PRG], NA, _, G, [T|NG]) :-
    T = c(NA2, _, _, _),    
    filter_group(PRG, NA, NA2, G, NG).

canonicalize((H :- G | B), N/A, Args, Guards, B2) :-
    expand_arg_pairs(H, B, H2, B2),
    H2 =.. [N|Args],
    length(Args, A),
    serialize_guards(G, Guards).
canonicalize((H :- B), N/A, Args, Guards, B2) :-
    expand_arg_pairs(H, B, H2, B1),
    canonicalize((H2 :- true | B1), N/A, Args, Guards, B2).
canonicalize(H, N/A, Args, Guards, B) :-
    expand_arg_pairs(H, true, H2, B1),
    canonicalize((H2 :- true | B1), N/A, Args, Guards, B).

serialize_guards(true, []).
serialize_guards((true, G1), G) :- serialize_guards(G1, G).
serialize_guards((G1, G2), [G1|G3]) :- serialize_guards(G2, G3).
serialize_guards(G, [G]).

write_translation([], _).
write_translation([[]|Gs], Out) :-
    write_translation(Gs, Out).
write_translation([G|Gs], Out) :-
    write_group(G, Out),
    nl(Out),
    write_translation(Gs, Out).

write_group([], _).
write_group([c(N/_, Args, _, B)|G], Out) :-
    H =.. [N|Args],
    format(Out, '~q.\n', [(H :- B)]),
    write_group(G, Out).

% translation chain

translate(G, TG) :-
    check_definitions(G),
    check_overlap(G),
    inject_equality_guards(G, G2),
    expand_terms(G2, G3),
    eliminate_guards(G3, TG).

% check for library redefinitions

check_definitions(Gs) :-
    member(G, Gs), 
    check_definitions2(G), 
    fail.
check_definitions(_).

check_definitions2([c(N/A, _, _, _)|_]) :-
    memberchk(N, ['+', '-', ',', ';', '|', ':']),
    error('redefinition of built-in operator ~a/~d\n', [N, A]).
check_definitions2(_).

% check argument overlap

check_overlap([]).
check_overlap([G|Gs]) :-
    check_group_overlap(G, 1, G),
    check_overlap(Gs).

check_group_overlap([], _, _).
check_group_overlap([c(_, [], _, _)|_], _, _) :- !.
check_group_overlap([c(_, Args, [], _)|_], I, G) :-
    find_overlap(G, Args, I, 1),
    !.
check_group_overlap([_|Cs], I, G) :-
    I2 is I + 1,
    check_group_overlap(Cs, I2, G).

find_overlap([], _, _, _) :- fail.
find_overlap([c(NA, Args1, [], _)|_], Args, I, J) :-
    I \= J,
    find_arg_overlap(Args1, Args, 1, 0, NA, I, J).
find_overlap([_|Cs], Args, I, J) :-
    J2 is J + 1,
    find_overlap(Cs, Args, I, J2).

find_arg_overlap([], _, _, 0, _, _, _) :- fail.
find_arg_overlap([], _, _, P, N/A, I, J) :- 
    P \= 0,
    format(user_error, 'Warning: clauses ~d and ~d of ~a/~d have overlapping match for arguments at position ~d but lack \'otherwise\' guard\n',
        [I, J, N, A, P]).
find_arg_overlap([X|_], [Y|_], _, _, _, _, _) :-
    discrete_value(X), 
    discrete_value(Y), 
    !, fail.
find_arg_overlap([X|_], [Y|_], _, _, _, _, _) :-
    \+congruent(X, Y),
    !, fail.
find_arg_overlap([X|Args1], [Y|Args2], P, 0, NA, I, J) :-
    structurally_same(X, Y),
    !,
    find_arg_overlap(Args1, Args2, P, P, NA, I, J).
find_arg_overlap([_|Args1], [_|Args2], P, O, NA, I, J) :-
    P2 is P + 1,
    find_arg_overlap(Args1, Args2, P2, O, NA, I, J).

discrete_value(X) :- atomic(X), !.
discrete_value('!'(_)) :- !.
discrete_value('$VAR'(_)) :- !, fail.
discrete_value([_|_]) :- !, fail.
discrete_value(X) :- X =.. [_, _|_].

congruent(X, _) :- atomic(X), !, fail.
congruent(_, Y) :- atomic(Y), !, fail.
congruent('$VAR'(_), _) :- !.
congruent(_, '$VAR'(_)) :- !.
congruent('!'(_), _) :- !.
congruent(_, '!'(_)) :- !.
congruent([H|T], Y) :-
    !,
    Y = [H2|T2],
    congruent(H, H2),
    congruent(T, T2).
congruent(X, Y) :- 
    X =.. [H|X2],
    Y =.. [H|Y2],
    congruent(X2, Y2).

structurally_same(X, X) :- !.
structurally_same('$VAR'(_), _) :- !.
structurally_same(_, '$VAR'(_)) :- !.
structurally_same(X, _) :- atomic(X), !, fail.
structurally_same(_, Y) :- atomic(Y), !, fail.
structurally_same([X1|X2], [Y1|Y2]) :-
    !,
    structurally_same(X1, Y1),
    structurally_same(X2, Y2).
structurally_same(X, Y) :-
    X =.. [HX|TX],
    Y =.. [HY|TY],
    structurally_same(HX, HY),
    structurally_same(TX, TY).
structurally_same(_, _).

% inject ==/2 guards for variables occurring multiple times in head

inject_equality_guards([], []).
inject_equality_guards([G1|Gs1], [G|Gs]) :-
    inject_equality_guards2(G1, G),
    inject_equality_guards(Gs1, Gs).

inject_equality_guards2([], []).
inject_equality_guards2([c(NA, Args1, Guards1, Body)|G1], [T|G]) :-
    T = c(NA, Args, Guards, Body),
    extract_variables((Args1, Body), Vars),
    duplicated_variables(Args1, Args, [], _, Vars, _, Guards1, Guards),
    inject_equality_guards2(G1, G).

duplicated_variables([], [], S, S, Vs, Vs, G, G).
duplicated_variables([Arg|Args], [Arg2|Args2], S1, S, Vs1, Vs, G1, G) :-
    duplicated_variables2(Arg, Arg2, S1, S2, Vs1, Vs2, G1, G2),
    duplicated_variables(Args, Args2, S2, S, Vs2, Vs, G2, G).

duplicated_variables2(X, X, S, S, Vs, Vs, G, G) :- atomic(X).
duplicated_variables2('$VAR'(V), '$VAR'(V2), S, S, Vs, [V2|Vs], G, [G1|G]) :-
    memberchk(V, S),
    available_variable(Vs, V2),
    G1 = ('$VAR'(V) == '$VAR'(V2)).
duplicated_variables2('$VAR'(V), '$VAR'(V), S, [V|S], Vs, Vs, G, G).
duplicated_variables2(X, '$VAR'(NV), S, S, Vs, [NV|Vs], G, [G1|G]) :-
    ground_value(X),
    available_variable(Vs, NV),
    G1 = ('$VAR'(NV) == X).
duplicated_variables2([X1|Y1], [X|Y], S1, S, Vs1, Vs, G1, G) :-
    duplicated_variables2(X1, X, S1, S2, Vs1, Vs2, G1, G2),
    duplicated_variables2(Y1, Y, S2, S, Vs2, Vs, G2, G).
duplicated_variables2(T1, T, S1, S, Vs1, Vs, G1, G) :-
    callable(T1),
    T1 =.. [Name|Args1],
    duplicated_variables(Args1, Args, S1, S, Vs1, Vs, G1, G),
    T =.. [Name|Args].
duplicated_variables2(X, X, S, S, Vs, Vs, G, G).

% eliminate guards by introducing new groups

eliminate_guards([], []).
eliminate_guards([G1|Gs1], [G|Gs]) :-
    eliminate_guards1(G1, G),
    eliminate_guards(Gs1, Gs).

eliminate_guards1([c(NA, Args, Guards1, B)|G1], G) :-
    % 'otherwise' in first clause has no effect
    delete(Guards1, otherwise, Guards),
    eliminate_guards2([c(NA, Args, Guards, B)|G1], G).

eliminate_guards2([], []).
eliminate_guards2([C|Cs1], [C|Cs]) :-
    C = c(_, _, [], _),
    % no guards, try next clause in group
    eliminate_guards2(Cs1, Cs).
eliminate_guards2([C1|Cs1], Cs) :-
    C1 = c(NA, Args, Guards, B),
    memberchk(otherwise, Guards),
    delete(Guards, otherwise, Guards2),
    eliminate_guards2([c(NA, Args, Guards2, B)|Cs1], [c(_, Args2, _, B2)|Cs3]),
    Cs = [c(NA, Args2, [], ('$suspend', B2))|Cs3].
eliminate_guards2([C1|Cs1], Cs) :-
    C1 = c(NA, Args, [data('$VAR'(V))|Guards], B),
    % data/1 guard
    force_variables(Args, [V], Args2),
    eliminate_guards2([c(NA, Args2, Guards, B)|Cs1], Cs).
eliminate_guards2([C1|Cs1], [C|Cs]) :-
    % clause with guards, excluding 'otherwise'
    C1 = c(NA, Args, Guards, B),
    C = c(NA, Args2, [], B2),
    extract_variables((Args, Guards, B), Vars),
    eliminate_guards3(Guards, Vars, Args, Args2, B, B2),
    eliminate_guards2(Cs1, Cs).

eliminate_guards3([], _, Args, Args, B, B).
eliminate_guards3([Guard|Guards], AllVars, Args, Args2, B1, B) :-
    extract_variables(Guard, HeadVars),
    available_variable(AllVars, HV),
    ResultVar = '$VAR'(HV),
    translate_guard(Guard, ResultVar, XGuard, [HV|AllVars], _, Force, Result),
    B = (XGuard, '$guard'(ResultVar, Result), B3),
    force_variables(Force, Args, HeadVars, Args3),
    eliminate_guards3(Guards, [HV|AllVars], Args3, Args2, B1, B3).

% utilities

rename_functor(N/A, P, Grist, RN) :-
    format_to_atom(RN, '$~d$~a$~a$~d', [P, Grist, N, A]).

extract_variables(X, Vs) :- extract_variables(X, Vs, []).

extract_variables(X, Vs, Vs) :- atomic(X).
extract_variables('$VAR'(V), [V|T], T).
extract_variables([X|Y], Vs, T) :-
    extract_variables(X, Vs, Vs2),
    extract_variables(Y, Vs2, T).
extract_variables(X, Vs, T) :-
    callable(X),
    X =.. [_|Args],
    extract_variables(Args, Vs, T).

force_variables(yes, X1, Vars, X) :- force_variables(X1, Vars, X).
force_variables(no, X, _, X).

force_variables(X, _, X) :- atomic(X).
force_variables('!'(X), _, '!'(X)).
force_variables('$VAR'(V), Vars, '!'('$VAR'(V))) :- memberchk(V, Vars), !.
force_variables('$VAR'(V), _, '$VAR'(V)).
force_variables([X|Y], Vars, [X2|Y2]) :- 
    force_variables(X, Vars, X2),
    force_variables(Y, Vars, Y2).
force_variables(X, Vars, X2) :-
    callable(X),
    X =.. [N|Args],
    force_variables(Args, Vars, Args2),
    X2 =.. [N|Args2].

available_variable(Vars, V) :- available_variables(0, Vars, [V]).

available_variables(Vars, Vs) :- available_variables(0, Vars, Vs).

available_variables(_, _, []).
available_variables(V, Vars, [V|L]) :- 
    \+memberchk(V, Vars), 
    !,
    V2 is V + 1,
    available_variables(V2, Vars, L).
available_variables(V1, Vars, L) :-
    V2 is V1 + 1,
    available_variables(V2, Vars, L).

ground_value(X) :- atomic(X), !.
ground_value([X|Y]) :- !, ground_value(X), ground_value(Y).
ground_value('$VAR'(_)) :- !, fail.
ground_value(T) :- 
    T =.. [_|Lst],
    ground_value(Lst).

% Guard translation

translate_guard(Exp, R, (X2, Y2, T), HV1, HV, Force, Result) :-
    callable(Exp),
    functor(Exp, Op, Arity),
    numeric_guard(Op/Arity),
    Exp =.. [_, X, Y],
    translate_expression(X, HV1, X2, XR, HV2),
    translate_expression(Y, HV2, Y2, YR, HV),
    Exp2 =.. [Op, XR, YR],
    translate_guard2(Exp2, R, T, Force, Result).
translate_guard(Exp, R, T, HV, HV, Force, Result) :-
    translate_guard2(Exp, R, T, Force, Result).

translate_guard2((X =:= Y), R, compute(=, X, Y, R), yes, true).
translate_guard2('\\=:='(X, Y), R, compute(=, X, Y, R), yes, false).
translate_guard2((X == Y), R, '$match'(X, Y, R), yes, true).
translate_guard2((X =\= Y), R, '$match'(X, Y, R), yes, false).
translate_guard2((X > Y), R, compute(>, X, Y, R), yes, true).
translate_guard2((X < Y), R, compute(<, X, Y, R), yes, true).
translate_guard2((X >= Y), R, compute(<, X, Y, R), yes, false).
translate_guard2((X =< Y), R, compute(>, X, Y, R), yes, false).
translate_guard2((X @< Y), R, '$below'(X, Y, R), yes, true).
translate_guard2((X @> Y), R, '$above'(X, Y, R), yes, true).
translate_guard2((X @>= Y), R, '$below'(Y, X, R), yes, true).
translate_guard2((X @=< Y), R, '$above'(Y, X, R), yes, true).
translate_guard2(idle, R, '$idle'(R), no, true).
translate_guard2(number(X), R, '$number'(X, R), yes, true).
translate_guard2(string(X), R, '$atom'(X, R), yes, true).
translate_guard2(atomic(X), R, '$atomic'(X, R), yes, true).
translate_guard2(list(X), R, '$list'(X, R), yes, true).
translate_guard2(tuple(X), R, '$tuple'(X, R), yes, true).
translate_guard2(unknown(X), R, '$var'(X, R), no, true).
translate_guard2(known(X), R, '$var'(X, R), no, false).
translate_guard2(G, _, _, _, _) :- error('invalid guard: ~q\n', [G]).

translate_expression(Exp, HV, true, Exp, HV) :- integer(Exp).
translate_expression('$VAR'(V), HV, true, '$VAR'(V), HV).
translate_expression(-X, HV1, X2, XR, HV) :-
    translate_expression(0-X, HV1, X2, XR, HV).
translate_expression(+X, HV1, X2, XR, HV) :-
    translate_expression(X, HV1, X2, XR, HV).
translate_expression(\X, HV1, X2, XR, HV) :-
    translate_expression('><'(-1, X), HV1, X2, XR, HV).
translate_expression(address_of(L), HV1, '$address_of'(L, '$VAR'(V)), '$VAR'(V), [V|HV1]) :-
    atom(L),
    available_variable(HV1, V).
translate_expression(Exp, HV1, (TX, TY, compute(NOp, RX, RY, '$VAR'(V))), '$VAR'(V), [V|HV3]) :-
    Exp =.. [Op, X, Y],
    numeric_operation(Op, NOp),
    translate_expression(X, HV1, TX, RX, HV2),
    translate_expression(Y, HV2, TY, RY, HV3),
    available_variable(HV3, V).
translate_expression(Exp, HV1, (TX, compute(NOp, RX, '$VAR'(V))), '$VAR'(V), [V|HV2]) :-
    Exp =.. [Op, X],
    numeric_operation(Op, NOp),
    translate_expression(X, HV1, TX, RX, HV2),
    available_variable(HV2, V).
translate_expression(Exp, _, _, _, _) :-
    error('invalid expression: ~w\n', [Exp]).

% expand is/2 expressions, explicit module calls and builtins:

expand_terms([], []).
expand_terms([G1|Gs1], [G|Gs]) :-
    expand_terms2(G1, G, Gs1, Gs2),
    expand_terms(Gs2, Gs).

expand_terms2([], [], Gs, Gs).
expand_terms2([C1|G1], [c(NA, Args, Guards, Body)|G], Gs1, Gs) :-
    C1 = c(NA, Args, Guards, Body1),
    extract_variables((Args, Guards, Body1), Vars),
    expand_body(Body1, Body, NA, Vars, _, Gs1, Gs2),
    expand_terms2(G1, G, Gs2, Gs).

counter(N) :-
    g_read(counter, N1),
    N is N1 + 1, 
    g_assign(counter, N).

%expand_body(X, _, _, _, _, _, _) :- dbg(X), fail.
expand_body((true, X), X2, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(X, X2, NA, Vars1, Vars, Gs1, Gs).
expand_body((X, true), X2, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(X, X2, NA, Vars1, Vars, Gs1, Gs).
expand_body(X = Y, '$unify_checked'(X, Y), _, Vs, Vs, Gs, Gs) :-
    extract_variables(X, Xvs),
    extract_variables(Y, Yvs),
    check_ref(Xvs, Yvs, (X = Y)).
expand_body(X + Y, X2, NA, Vars1, Vars, Gs1, Gs) :-
    expand_arg_pairs(X + Y, no, [], Vars1, X1, _, Vars2),
    expand_body(X1, X2, NA, Vars2, Vars, Gs1, Gs).
expand_body(':='(X, Y), '$assign'(Y, X), _, Vs, Vs, Gs, Gs) :-
    check_var(X, ':='(X, Y)),
    X = '$VAR'(AV),
    extract_variables(Y, Vars),
    check_ref([AV], Vars, ':='(X, Y)).
expand_body(X is Exp, T, NA, Vars, [RV|Vars], Gs1, [G|Gs1]) :-
    check_var(X, (X is Exp)),
    counter(C),
    rename_functor(NA, C, x, N2),
    extract_variables(Exp, ExpVars),
    X = '$VAR'(AV),
    check_ref([AV], ExpVars, (X is Exp)),
    sort(ExpVars, SortedVars),
    length(SortedVars, SL),
    available_variable(Vars, RV),
    translate_expression(Exp, [RV|SortedVars], XBody, Result, _),
    build_exp_args(SortedVars, Args, Params, ['$VAR'(RV)]),
    Call =.. [N2|Params],
    T = (Call, '$assign'('$VAR'(RV), X)),
    SL2 is SL + 1,
    assertz(declaration(synthesized(N2/SL2, NA))),
    G = [c(N2/SL2, Args, [], (XBody, '$assign'(Result, '$VAR'(RV))))].
expand_body(when(X, T), T2, NA, Vars, Vars, Gs1, [G|Gs1]) :-
    check_when(X, T, V),
    counter(C),
    rename_functor(NA, C, w, N2),
    extract_variables((X, T), WVars),
    sort(WVars, SortedVars),
    length(SortedVars, SL),
    build_when_args(SortedVars, V, Args, Params),
    T2 =.. [N2|Params],
    assertz(declaration(synthesized(N2/SL, NA))),
    G = [c(N2/SL, Args, [], T)].
expand_body(';'('->'(X, Y), Z), Call, NA, Vars, Vars, Gs1, [G|Gs1]) :-
    extract_variables((X, Y, Z), AllVars),
    sort(AllVars, SortedVars),
    length(SortedVars, SL),
    counter(C),
    rename_functor(NA, C, c, N2),
    varlist(SortedVars, CArgs),
    serialize_guards(X, XG),
    Call =.. [N2|CArgs],
    assertz(declaration(synthesized(N2/SL, NA))),
    G = [c(N2/SL, CArgs, XG, Y), 
         c(N2/SL, CArgs, [otherwise], Z)].
expand_body('->'(X, Y), T2, NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(';'('->'(X, Y), true), T2, NA, Vars1, Vars, Gs1, Gs).
expand_body(format(Fstr, Args), X, NA, Vars1, Vars, Gs1, Gs) :-
    compile_format(Fstr, Args, NA, Items),
    available_variable(Vars1, RV),
    C = format_chunked(Items, '$VAR'(RV)),
    expand_body(C, X, NA, [RV|Vars1], Vars, Gs1, Gs).
expand_body(format(Fstr, Args, Done), X, NA, Vars1, Vars, Gs1, Gs) :-
    compile_format(Fstr, Args, NA, Items),
    available_variable(Vars1, RV),
    C = format_chunked(Items, Done),
    expand_body(C, X, NA, [RV|Vars1], Vars, Gs1, Gs).
expand_body(format_chars(Fstr, Args, Out, Tail), X, NA, Vars1, Vars, Gs1, Gs) :-
    compile_format(Fstr, Args, NA, Items),
    C = '$outputs'(Items, Out, Tail),
    expand_body(C, X, NA, Vars1, Vars, Gs1, Gs).
expand_body(T, X, NA, Vars1, Vars, Gs1, Gs) :-
    functor(T, N, A),
    applicative(N/A),
    T =.. [_, Goal|Args],
    canonicalize_call_target(Goal, Goal2, Gs1, Gs2),
    T2 =.. [N, Goal2|Args],
    expand_body(T2, X, NA, Vars1, Vars, Gs2, Gs).    
expand_body(T, X, NA, Vars1, Vars, Gs1, Gs) :-
    functor(T, N, A),
    applicative(N/A),
    T =.. [_, Goal|Args],
    local_call(Goal),
    expand_applicative(N, Goal, Args, X, NA, Vars1, Vars, Gs1, Gs).
expand_body((X, Y), (X2, Y2), NA, Vars1, Vars, Gs1, Gs) :-
    expand_body(X, X2, NA, Vars1, Vars2, Gs1, Gs2),
    expand_body(Y, Y2, NA, Vars2, Vars, Gs2, Gs).
expand_body(T, T, _, Vs, Vs, Gs, Gs).

% convert goal to callable user goal or fully qualified library call, or fail:
canonicalize_call_target(Goal1, Goal, Gs1, Gs) :-
    primitive_call(Goal1),
    !,
    wrap_primitive(Goal1, Goal, Gs1, Gs).

wrap_primitive(T, T2, Gs, [G|Gs]) :-
    counter(C),
    T =.. [N|Args],
    length(Args, A),
    extract_variables(T, Vars),
    sort(Vars, SortedVars),
    length(SortedVars, SL),
    rename_functor(N/A, C, p, N2),
    var_lists(SortedVars, [], Args2, Params),
    T2 =.. [N2|Args2],
    Call =.. [N|Args],
    assertz(declaration(synthesized(N2/SL, N/A))),
    G = [c(N2/SL, Params, [], Call)].

wrap_term(T, NA, T2, Gs, [G|Gs]) :-
    counter(C),
    extract_variables(T, Vars),
    sort(Vars, SortedVars),
    length(SortedVars, SL),
    rename_functor(NA, C, t, N2),
    varlist(SortedVars, Args2),
    T2 =.. [N2|Args2],
    assertz(declaration(synthesized(N2/SL, NA))),
    G = [c(N2/SL, Args2, [], T)].

local_call(T) :-
    callable(T),
    T \= '$VAR'(_),
    T \= _:_,
    \+primitive_call(T).

check_var('$VAR'(_), _) :- !.
check_var(_, X) :- error('expected variable in assignment: ~q\n', [X]).

check_ref(Vs1, Vs2, X) :- 
    intersect(Vs1, Vs2, Vs),
    Vs \= [],
    error('circular reference in binding: ~q\n', [X]).
check_ref(_, _, _).

check_when('$VAR'(V), _, V).
check_when(X, T, _) :- error('invalid argument to `when\': ~q\n', [when(X, T)]).

check_module(M, _) :- atom(M), !.
check_module(_, X) :- error('invalid module prefix in `~q\'\n', [X]).

intersect([], _, []).
intersect([X|L1], L2, [X|LI]) :-
    memberchk(X, L2),
    intersect(L1, L2, LI).
intersect([_|L1], L2, LI) :- intersect(L1, L2, LI).

build_when_args([], _, [], []).
build_when_args([V|Vs], V, ['!'('$VAR'(V))|Args], ['$VAR'(V)|Params]) :-
    build_when_args(Vs, V, Args, Params).
build_when_args([V1|Vs], V, ['$VAR'(V1)|Args], ['$VAR'(V1)|Params]) :-
    build_when_args(Vs, V, Args, Params).

expand_builtin_call(simple, Args, Op, _, R, Gs, Gs) :-
    T =.. [Op|Args],
    R = foreign_call(T).
expand_builtin_call(Spec, Args, Op, NA, R, Gs, [G|Gs]) :-
    counter(C),
    rename_functor(NA, C, f, N2),
    extract_variables(Args, Vars),
    sort(Vars, SortedVars),
    length(SortedVars, SL),
    scan_builtin_args(Spec, Args, Forced),
    var_lists(SortedVars, Forced, Args2, Params),
    T =.. [Op|Args],
    R =.. [N2|Params],
    assertz(declaration(synthesized(N2/SL, NA))),
    G = [c(N2/SL, Args2, [], foreign_call(T))].

scan_builtin_args([], _, []).
scan_builtin_args([f|Spec], ['$VAR'(V)|Args], [V|Forced]) :-
    scan_builtin_args(Spec, Args, Forced).
scan_builtin_args([_|Spec], [_|Args], Forced) :-
    scan_builtin_args(Spec, Args, Forced).

varlist([], []).
varlist([V|L], ['$VAR'(V)|VL]) :- varlist(L, VL).

var_lists([], _, [], []).
var_lists([V|Vs], Forced, ['!'('$VAR'(V))|Args], ['$VAR'(V)|Params]) :-
    member(V, Forced),
    var_lists(Vs, Forced, Args, Params).
var_lists([V|Vs], Forced, ['$VAR'(V)|Args], ['$VAR'(V)|Params]) :-
    var_lists(Vs, Forced, Args, Params).

build_exp_args([], T, T, T).
build_exp_args([V|Vs], ['!'('$VAR'(V))|XVs], ['$VAR'(V)|PVs], T) :-
    build_exp_args(Vs, XVs, PVs, T).

% precompilation of format strings

compile_format([], [], _, []).
compile_format([], Args, N/A, []) :-
    warning('~a/~d: extra arguments in format string: ~w\n', [N, A, Args]).
compile_format(Fstr, Args, NA, Out) :-
    atom(Fstr), 
    atom_codes(Fstr, Str),
    compile_format(Str, Args, NA, Out).
compile_format(Fstr, _, _, _) :-
    \+ground_format_string(Fstr),
    !, fail.
compile_format(_, Args, _, _) :-
    \+valid_format_args(Args),
    !, fail.
compile_format([0'~, 0'~|Fstr], Args, NA, [0'~|Out]) :-
    compile_format(Fstr, Args, NA, Out).
compile_format([0'~, 0'n|Fstr], Args, NA, [10|Out]) :-
    compile_format(Fstr, Args, NA, Out).
compile_format([0'~, C|Fstr], Args,NA,  Out) :-
    compile_format_char(C, Fstr, Args, NA, Out).
compile_format(Fstr, Args, NA, [S|Out]) :-
    compile_format_chunk(Fstr, Fstr2, CL),
    atom_codes(S, CL),
    compile_format(Fstr2, Args, NA, Out).

compile_format_char(C, Fstr, [], N/A, [0'~, C|Out]) :-
    warning('~a/~d: missing argument for ~~~c in format string: ...~s\n', 
        [N, A, C, Fstr]),
    compile_format(Fstr, [], N/A, Out).
compile_format_char(0'c, Fstr, [X|Args], NA, [X|Out]) :-
    compile_format(Fstr, Args, NA, Out).
compile_format_char(0'?, Fstr, [F, Args1|Args], NA, [f(F, Args1)|Out]) :-
    compile_format(Fstr, Args, NA, Out).
compile_format_char(0'?, Fstr, _, N/A, [0'~, 0'?|Out]) :-
    warning('~a/~d: missing argument for ~~? in format string: ...~s\n', 
        [N, A, Fstr]),
    compile_format(Fstr, [], N/A, Out).
compile_format_char(C, Fstr, [X|Args], NA, [Item|Out]) :-
    member(C, [0'd, 0'x, 0'a, 0's, 0'q, 0'w]),
    atom_codes(N, [C]),
    Item =.. [N, X],
    compile_format(Fstr, Args, NA, Out).
compile_format_char(C, Fstr, Args, N/A, [0'~, C|Out]) :-
    warning('~a/~d: invalid formatting directive in format string: ...~s\n', 
        [N, A, [C|Fstr]]),
    compile_format(Fstr, Args, N/A, Out).

compile_format_chunk([], [], []).
compile_format_chunk([0'~|Fstr], [0'~|Fstr], []).    
compile_format_chunk([C|Fstr1], Fstr, [C|S]) :-
    compile_format_chunk(Fstr1, Fstr, S).

ground_format_string([]).
ground_format_string([C|X]) :-
    integer(C), 
    ground_format_string(X).

valid_format_args([]).
valid_format_args([_|Args]) :- valid_format_args(Args).

% expanding argument pairs

expand_arg_pairs(H, B, H2, B2) :-
    \+atom(H),
    extract_variables((H, B), Vars),
    extract_head_pairs(H, [], Vars, H2, Pairs, Vars2),
    expand_arg_pairs(B, yes, Pairs, Vars2, B1, Pairs2, _),
    finalize_arg_pairs(Pairs2, B1, B2).
expand_arg_pairs(H, B, H, B).

extract_head_pairs((H-'$VAR'(P)), Pairs1, Vars1, H2, Pairs, Vars) :-
    extract_head_pairs(H, Pairs1, Vars1, H1, Pairs2, Vars2),
    H1 =.. [N|Args],
    available_variable(Vars2, P1),
    append(Args, ['$VAR'(P), '$VAR'(P1)], Args2),
    H2 =.. [N|Args2],
    Pairs = [P-(P/P1)|Pairs2],
    Vars = [P1|Vars2].
extract_head_pairs((H+X), Pairs1, Vars1, H2, Pairs, Vars) :-
    extract_head_pairs(H, Pairs1, Vars1, H1, Pairs, Vars),
    H1 =.. [N|Args],
    append(Args, [X], Args2),
    H2 =.. [N|Args2].
extract_head_pairs(H, Pairs, Vars, H, Pairs, Vars).

expand_arg_pairs((T1, T2), Tail, Pairs1, Vars1, (T1b, T2b), Pairs, Vars) :-
    expand_arg_pairs(T1, no, Pairs1, Vars1, T1b, Pairs2, Vars2),
    expand_arg_pairs(T2, Tail, Pairs2, Vars2, T2b, Pairs, Vars).
expand_arg_pairs((T1, T2), Tail, Pairs1, Vars1, (T1b, T2b), Pairs, Vars) :-
    expand_arg_pairs(T1, no, Pairs1, Vars1, T1b, Pairs2, Vars2),
    expand_arg_pairs(T2, Tail, Pairs2, Vars2, T2b, Pairs, Vars).
expand_arg_pairs(when(X, T), Tail, Pairs1, Vars1, when(X2, T2), Pairs, Vars) :-
    expand_pair_val(X, Pairs1, X2),
    expand_arg_pairs(T, Tail, Pairs1, Vars1, T2, Pairs, Vars).
expand_arg_pairs(';'('->'(X, T1), T2), Tail, Pairs1, Vars1, R, Pairs, Vars) :-
    expand_pair_val(X, Pairs1, X2),
    expand_arg_pairs(T1, Tail, Pairs1, Vars1, T1b, Pairs2, Vars2),
    expand_arg_pairs(T2, Tail, Pairs1, Vars2, T2b, Pairs3, Vars3),
    combine_pairs(Pairs2, Pairs3, C1, C2, Pairs, Vars3, Vars),
    R = (';'('->'(X2, (T1b, C1)), (T2b, C2))).
expand_arg_pairs('->'(X, T), Tail, Pairs1, Vars1, T2, Pairs, Vars) :-
    expand_arg_pairs(';'('->'(X, T), true), Tail, Pairs1, Vars1, T2, Pairs, Vars).
expand_arg_pairs('<='('$VAR'(P), X), Tail, Pairs1, Vars1, T, Pairs, Vars) :-
    expand_pair_val(X, Pairs1, X2),
    next_pair(P, '<='('$VAR'(P), X), Tail, P1, P2, Vars1, Vars, Pairs1, Pairs),
    T = ('$VAR'(P1) = [X2|'$VAR'(P2)]).
expand_arg_pairs('=>'(X, '$VAR'(P)), Tail, Pairs1, Vars1, T, Pairs, Vars) :-
    expand_pair_val(X, Pairs1, X2),
    next_pair(P, '=>'(X, '$VAR'(P)), Tail, P1, P2, Vars1, Vars, Pairs1, Pairs),
    T = ([X2|'$VAR'(P1)] = '$VAR'(P2)).
expand_arg_pairs('<=='('$VAR'(P), X), Tail, Pairs1, Vars1, T, Pairs, Vars) :-
    expand_pair_val(X, Pairs1, X2),
    next_pair(P, '<=='('$VAR'(P), X), Tail, _, P2, Vars1, Vars, Pairs1, Pairs),
    T = ('$VAR'(P2) = X2).
expand_arg_pairs(XT, Tail, Pairs1, Vars1, T, Pairs, Vars) :-
    callable(XT),
    XT =.. [AOp, '$VAR'(P), X],
    member(a(AOp, Op), [a('+=', '+'), a('-=', '-'), a('*=', '*'), a('/=', '/')]),
    expand_pair_val(X, Pairs1, X2),
    next_pair(P, XT, Tail, P1, P2, Vars1, Vars, Pairs1, Pairs),
    Exp =.. [Op, '$VAR'(P1), X2],
    T = ('$VAR'(P2) is Exp).
expand_arg_pairs(AT, Tail, Pairs1, Vars1, AT2, Pairs, Vars) :-
    callable(AT),
    functor(AT, Op, A),
    (applicative(Op/A); Op/A = apply/2),
    AT =.. [_, T|Args],
    expand_arg_pairs(T, Tail, Pairs1, Vars1, T2, Pairs, Vars),
    AT2 =.. [Op, T2|Args].
expand_arg_pairs(T-'$VAR'(P), Tail, Pairs1, Vars1, T2, Pairs, Vars) :-
    callable(T),
    expand_arg_pairs(T, Tail, Pairs1, Vars1, T1, Pairs2, Vars2),
    T1 =.. [N|Args],
    next_pair(P, T-'$VAR'(P), Tail, P1, P2, Vars2, Vars, Pairs2, Pairs),
    append(Args, ['$VAR'(P1), '$VAR'(P2)], Args2),
    T2 =.. [N|Args2].
expand_arg_pairs(T+X, Tail, Pairs1, Vars1, T2, Pairs, Vars) :-
    callable(T),
    expand_arg_pairs(T, Tail, Pairs1, Vars1, T1, Pairs, Vars),
    T1 =.. [N|Args],
    append(Args, [X], Args2),
    T2 =.. [N|Args2].
expand_arg_pairs(T, _, Pairs, Vars, T, Pairs, Vars).

expand_pair_val(X, _, X) :- atomic(X).
expand_pair_val('$VAR'(P), Pairs, '$VAR'(P1)) :- member(P-(P1/_), Pairs).
expand_pair_val([X|Y], Pairs, [X2|Y2]) :-
    expand_pair_val(X, Pairs, X2),
    expand_pair_val(Y, Pairs, Y2).
expand_pair_val(X, Pairs, X2) :-
    callable(X),
    X =.. [N|Args],
    expand_pair_val(Args, Pairs, Args2),
    X2 =.. [N|Args2].
expand_pair_val(X, _, X).

next_pair(P, _, yes, P1, P2, Vars, Vars, Pairs1, Pairs) :-
    select(P-(P1/P2), Pairs1, Pairs).
next_pair(_, X, yes, _, _, _, _, _, _) :-
    error('reference to undefined argument pair: ~q\n', [X]).
next_pair(P, _, no, P1, P3, Vars, [P3|Vars], Pairs1, Pairs) :-
    select(P-(P1/P2), Pairs1, Pairs2),
    available_variable(Vars, P3),
    Pairs = [P-(P3/P2)|Pairs2].
next_pair(_, X, no, _, _, _, _, _, _) :-
    error('reference to undefined argument pair: ~q\n', [X]).

finalize_arg_pairs([], B, B).
finalize_arg_pairs([_-(P1/P2)|Pairs], B1, B) :-
    finalize_arg_pairs(Pairs, (B1, '$VAR'(P1) = '$VAR'(P2)), B).

combine_pairs([], Pairs, true, true, Pairs, Vars, Vars).
combine_pairs(Pairs, Pairs2, C1, C2, Pairs3, Vars1, Vars) :-
    select(P-(PA/P2), Pairs, PairsA),
    select(P-(PB/P2), Pairs2, PairsB),
    available_variable(Vars1, P3),
    C1 = ('$VAR'(P3) = '$VAR'(PA), CA),
    C2 = ('$VAR'(P3) = '$VAR'(PB), CB),
    Pairs3 = [P-(P3/P2)|PairsN],
    combine_pairs(PairsA, PairsB, CA, CB, PairsN, [P3|Vars1], Vars).
combine_pairs(Pairs, Pairs2, true, true, Pairs3, Vars, Vars) :-
    append(Pairs, Pairs2, Pairs3).

% expansion of applicative forms
%XXX refactor this

expand_applicative(foreach, Goal, [List, Done], X, NA, Vars, Vars, Gs, [G|Gs]) :-
    extract_variables((Goal, List, Done), AllVars),
    sort(AllVars, SortedVars),
    length(SortedVars, A),
    ALoop is A + 2,
    counter(C),
    rename_functor(NA, C, a, NLoop),
    varlist(SortedVars, CArgs),
    available_variables(AllVars, [LV, XV, DV]),
    assertz(declaration(synthesized(NLoop/ALoop, NA))),
    varlist([LV, XV, DV], [LVar, XVar, DVar]),
    X =.. [NLoop, List, Done|CArgs],
    Goal =.. [NGoal|GArgs],
    append(GArgs, [XVar], AArgs),
    Action =.. [NGoal|AArgs],
    Repeat =.. [NLoop, LVar, DVar|CArgs],
    G = [c(NLoop/ALoop, [[], DVar|CArgs], [], '$assign'([], DVar)), 
         c(NLoop/ALoop, [[XVar|LVar], DVar|CArgs], [], (Action, Repeat))].
expand_applicative(maplist, Goal, [List, Result], X, NA, Vars, Vars, Gs, [G|Gs]) :-
    extract_variables((Goal, List, Result), AllVars),
    sort(AllVars, SortedVars),
    length(SortedVars, A),
    ALoop is A + 2,
    counter(C),
    rename_functor(NA, C, a, NLoop),
    varlist(SortedVars, CArgs),
    available_variables(AllVars, [LV, RV, XV, VV, R2V]),
    assertz(declaration(synthesized(NLoop/ALoop, NA))),
    varlist([LV, RV, XV, VV, R2V], [LVar, RVar, XVar, VVar, R2Var]),
    X =.. [NLoop, List, Result|CArgs],
    Goal =.. [NGoal|GArgs],
    append(GArgs, [XVar, VVar], AArgs),
    Action =.. [NGoal|AArgs],
    Repeat =.. [NLoop, LVar, R2Var|CArgs],
    G = [c(NLoop/ALoop, [[], RVar|CArgs], [], '$assign'([], RVar)), 
         c(NLoop/ALoop, [[XVar|LVar], RVar|CArgs], [], 
            (Action, RVar = [VVar|R2Var], Repeat))].
expand_applicative(filter, Goal, [List, Result], X, NA, Vars, Vars, Gs, [G|Gs]) :-
    extract_variables((Goal, List, Result), AllVars),
    sort(AllVars, SortedVars),
    length(SortedVars, A),
    ALoop is A + 2,
    counter(C),
    rename_functor(NA, C, a, NLoop),
    varlist(SortedVars, CArgs),
    available_variables(AllVars, [LV, RV, XV, VV, R2V]),
    assertz(declaration(synthesized(NLoop/ALoop, NA))),
    varlist([LV, RV, XV, VV, R2V], [LVar, RVar, XVar, VVar, R2Var]),
    X =.. [NLoop, List, Result|CArgs],
    Goal =.. [NGoal|GArgs],
    append(GArgs, [XVar, VVar], AArgs),
    Action =.. [NGoal|AArgs],
    RepeatF =.. [NLoop, LVar, RVar|CArgs],
    RepeatT =.. [NLoop, LVar, R2Var|CArgs],
    G = [c(NLoop/ALoop, [[], RVar|CArgs], [], '$assign'([], RVar)), 
         c(NLoop/ALoop, [[XVar|LVar], RVar|CArgs], [], 
            (Action, (VVar == false -> RepeatF; RVar = [XVar|R2Var], RepeatT)))].
expand_applicative(partition, Goal, [List, TResult, FResult], X, NA, Vars, Vars, Gs, 
    [G|Gs]) :-
    extract_variables((Goal, List, TResult, FResult), AllVars),
    sort(AllVars, SortedVars),
    length(SortedVars, A),
    ALoop is A + 3,
    counter(C),
    rename_functor(NA, C, a, NLoop),
    varlist(SortedVars, CArgs),
    available_variables(AllVars, [LV, TV, FV, XV, VV, TRV, FRV]),
    assertz(declaration(synthesized(NLoop/ALoop, NA))),
    varlist([LV, TV, FV, XV, VV, TRV, FRV], 
        [LVar, TVar, FVar, XVar, VVar, TRVar, FRVar]),
    X =.. [NLoop, List, TResult, FResult|CArgs],
    Goal =.. [NGoal|GArgs],
    append(GArgs, [XVar, VVar], AArgs),
    Action =.. [NGoal|AArgs],
    RepeatF =.. [NLoop, LVar, TVar, FRVar|CArgs],
    RepeatT =.. [NLoop, LVar, TRVar, FVar|CArgs],
    G = [c(NLoop/ALoop, [[], TVar, FVar|CArgs], [], ('$assign'([], TVar), 
          '$assign'([], FVar))), 
         c(NLoop/ALoop, [[XVar|LVar], TVar, FVar|CArgs], [], 
            (Action, (VVar == false -> FVar = [XVar|FRVar], RepeatF; 
                TVar = [XVar|TRVar], RepeatT)))].
expand_applicative(foldl, Goal, [List, Init, Result], X, NA, Vars, Vars, Gs, 
    [G|Gs]) :-
    extract_variables((Goal, List, Init, Result), AllVars),
    sort(AllVars, SortedVars),
    length(SortedVars, A),
    ALoop is A + 3,
    counter(C),
    rename_functor(NA, C, a, NLoop),
    varlist(SortedVars, CArgs),
    available_variables(AllVars, [LV, RV, IV, XV, VV]),
    assertz(declaration(synthesized(NLoop/ALoop, NA))),
    varlist([LV, RV, IV, XV, VV], [LVar, RVar, IVar, XVar, VVar]),
    X =.. [NLoop, List, Init, Result|CArgs],
    Goal =.. [NGoal|GArgs],
    append(GArgs, [IVar, XVar, VVar], AArgs),
    Action =.. [NGoal|AArgs],
    Repeat =.. [NLoop, LVar, VVar, RVar|CArgs],
    G = [c(NLoop/ALoop, [[], IVar, RVar|CArgs], [], RVar = IVar), 
         c(NLoop/ALoop, [[XVar|LVar], IVar, RVar|CArgs], [], 
            (Action, Repeat))].
expand_applicative(any, Goal, [List, Result], X, NA, Vars, Vars, Gs, [G|Gs]) :-
    extract_variables((Goal, List, Result), AllVars),
    sort(AllVars, SortedVars),
    length(SortedVars, A),
    ALoop is A + 2,
    counter(C),
    rename_functor(NA, C, a, NLoop),
    varlist(SortedVars, CArgs),
    available_variables(AllVars, [LV, RV, XV, VV]),
    assertz(declaration(synthesized(NLoop/ALoop, NA))),
    varlist([LV, RV, XV, VV], [LVar, RVar, XVar, VVar]),
    X =.. [NLoop, List, Result|CArgs],
    Goal =.. [NGoal|GArgs],
    append(GArgs, [XVar, VVar], AArgs),
    Action =.. [NGoal|AArgs],
    Repeat =.. [NLoop, LVar, RVar|CArgs],
    G = [c(NLoop/ALoop, [[], RVar|CArgs], [], '$assign'(false, RVar)), 
         c(NLoop/ALoop, [[XVar|LVar], RVar|CArgs], [], 
            (Action, (VVar == false -> Repeat; '$assign'(true, RVar))))].
expand_applicative(every, Goal, [List, Result], X, NA, Vars, Vars, Gs, [G|Gs]) :-
    extract_variables((Goal, List, Result), AllVars),
    sort(AllVars, SortedVars),
    length(SortedVars, A),
    ALoop is A + 2,
    counter(C),
    rename_functor(NA, C, a, NLoop),
    varlist(SortedVars, CArgs),
    available_variables(AllVars, [LV, RV, XV, VV]),
    assertz(declaration(synthesized(NLoop/ALoop, NA))),
    varlist([LV, RV, XV, VV], [LVar, RVar, XVar, VVar]),
    X =.. [NLoop, List, Result|CArgs],
    Goal =.. [NGoal|GArgs],
    append(GArgs, [XVar, VVar], AArgs),
    Action =.. [NGoal|AArgs],
    Repeat =.. [NLoop, LVar, RVar|CArgs],
    G = [c(NLoop/ALoop, [[], RVar|CArgs], [], '$assign'(true, RVar)), 
         c(NLoop/ALoop, [[XVar|LVar], RVar|CArgs], [], 
            (Action, (VVar == false -> '$assign'(false, RVar); Repeat)))].
expand_applicative(sequence, Goal, [Count, Result, Tail], X, NA, Vars, Vars, Gs, 
    [G|Gs]) :-
    extract_variables((Goal, Count, Result), AllVars),
    sort(AllVars, SortedVars),
    length(SortedVars, A),
    ALoop is A + 3,
    counter(C),
    rename_functor(NA, C, a, NLoop),
    varlist(SortedVars, CArgs),
    available_variables(AllVars, [IV, I2V, XV, RV, R2V, TV]),
    assertz(declaration(synthesized(NLoop/ALoop, NA))),
    varlist([IV, I2V, XV, RV, R2V, TV], [IVar, I2Var, XVar, RVar, R2Var, TVar]),
    X =.. [NLoop, 1, Result, Tail|CArgs],
    Goal =.. [NGoal|GArgs],
    append(GArgs, [IVar, XVar], AArgs),
    Action =.. [NGoal|AArgs],
    Repeat =.. [NLoop, I2Var, R2Var, TVar|CArgs],
    G = [c(NLoop/ALoop, [IVar, RVar, TVar|CArgs], [IVar > Count], '$assign'(TVar, RVar)), 
         c(NLoop/ALoop, [IVar, RVar, TVar|CArgs], [otherwise], 
            (Action, RVar = [XVar|R2Var], I2Var is IVar + 1, Repeat))].
expand_applicative(N, Goal, Args, X, _, Vars, Vars, Gs, Gs) :-
    X =.. [N, Goal|Args].

% tables

primitive_call(_=_).
primitive_call('$assign'(_, _)).
primitive_call('$error'(_)).
primitive_call('$dbg'(_, _)).
primitive_call(unify(_, _, _)).
primitive_call(compute(_, _, _)).
primitive_call(compute(_, _, _, _)).
primitive_call(foreign_call(_)).

numeric_guard(('=:=')/2).
numeric_guard(('\\=:=')/2).
numeric_guard(('>')/2).
numeric_guard(('<')/2).
numeric_guard(('>=')/2).
numeric_guard(('=<')/2).

numeric_operation('/\\', and).
numeric_operation('\\/', or).
numeric_operation('><', xor).
numeric_operation('<<', shl).
numeric_operation('>>', shr).
numeric_operation('\\\\', mod).
numeric_operation(Op, Op) :- numeric_operation(Op).

numeric_operation(+).
numeric_operation(-).
numeric_operation(*).
numeric_operation(/).
numeric_operation(sqrt).
numeric_operation(max).
numeric_operation(min).
numeric_operation(sign).
numeric_operation(abs).

applicative(maplist/3).
applicative(foreach/3).
applicative(filter/3).
applicative(partition/4).
applicative(foldl/4).
applicative(mapreduce/4).
applicative(any/3).
applicative(every/3).
applicative(sequence/4).
