; FLENG - 6502 runtime system (C64)
;
; register usage:
;
;   A: LO, HI
;   G: GOAL
;   N: NEWGOAL
;   P0-P3: PARAM[<I>]
;   E0-E<n>: ENV[<I>]
;
; encoding:
;
;   string:             aaaaaaaa aaaaaaa1
;   immediate integer:  0snnnnnn nnnnnnn1
;   cell:               1aaaaaaa aaaaaaa0
;
;   cell = [TAG, CAR, CDR]
;
;   tag:    _TLIVAAA cccccccc
;
;   T = 1: tuple ("AAA" is arity / length - 1)
;   V = 1: var
;   I = 1: integer
;   L = 1: list
;
;   List = [Car, Cdr]
;   Tuple = [Head, Arglist]
;   Integer = [Magnitude, _]
;   Var = [Val|Var, S-list]
;
;   S-List = [[Goal, [Var', Addr], ...], ...]
;
;   Var' = marked var (lowest bit set)
;
; type-ordering: int < string < list < tuple

IMMEDIATE_BIT = 0
LOW_BIT = 7

VAR_BIT = 3
INT_BIT = 4
LIST_BIT = 5
TUPLE_BIT = 6

TIMESLICE = 10

R6510 = $0001
CHROUT = $ffd2
PLOT = $fff0
STKEY = $91

TAG_MASK = $ff00
COUNT_MASK = $ff
ARITY_MASK = $0700
BITS_MASK = $f800

MAX_ENV     = 32
HEAP    = $8000
MAX_CELLS = $d00
MAX_GOALS = $400

; zeropage:
LO = $fb    ; FREKZP
HI = $fc
X0 = $fd    ; BASZPT
X1 = $22    ; INDEX1
AP = $24    ; INDEX2
GOAL = $26  ; RESHO
NGOAL = $28 ; RESMO
PARAM0 = $36 ; FRESPC
PARAM1 = $38
PARAM2 = $3a
PARAM3 = $3c
consing = $40
up1 = $42
up2 = $44
qstart = $46
qend = $48
sstackp = $4a

BASE = $801

    .word $801
    * = $801

    ; "10 SYS 2064"
    .byte $0C, $08, $0A, $00, $9E, $20, $32, $30, $36, $34, $00, $00, $00
    .byte 0, 0

    ; entry point
    tsx
    ;XXX simply set to top?
    stx stackbase
    ; disable BASIC ROM
    lda R6510
    and #254
    sta R6510
    jmp init

    ; extract integer value
    ; A=val -> A (int)
fl_getint:
    lda LO
    and #1 << IMMEDIATE_BIT
    beq l1
    sec
    bit HI
    bvs lint1
    lda HI          ; positive, clear LOW_BIT
    and #$7f
    sta HI
    clc
lint1:
    lda HI
    ror
    sta HI
    lda LO
    ror
    sta LO
    rts
l1:
    ldy #2
    lda (LO), y
    tax
    iny
    lda (LO), y
    stx LO
    sta HI
    rts

    ; create integer
    ; A=int -> A
fl_mkint:
    lda HI
    and #$c0
    beq lmki1
    cmp #$c0
    bne l4
lmki1:
    lda LO        ; immediate integer
    sec
    rol
    sta LO
    lda HI
    rol
    ora #1 << LOW_BIT
    sta HI
    rts
l4:
    lda #0
    sta X0
    sta X0 + 1
fl_mkbigint:
    lda #1 << INT_BIT
    ; fall through...
    
    ; allocate cell
    ; A=car, X0=cdr, a=tag -> A
fl_cons: 
    pha
    lda freelist
    ora freelist + 1
    bne l6
    lda #<ENOCELLS
    sta X0
    lda #>ENOCELLS
    sta X0 + 1
    jmp error
l6:
    lda freelist
    sta consing
    lda freelist + 1
    sta consing + 1
    ldy #0
    lda (consing), y
    sta freelist
    iny
    lda (consing), y
    sta freelist + 1
    pla
    sta (consing), y
    dey
    lda #0
    sta (consing), y
    jsr fl_addref
    lda LO
    ldy #2
    sta (consing), y
    lda HI
    iny
    sta (consing), y
    lda X0
    sta LO
    lda X0 + 1
    sta HI
    jsr fl_addref
    lda LO
    ldy #4
    sta (consing), y
    lda HI
    iny
    sta (consing), y
    lda cells
    clc
    adc #1
    sta cells
    lda cells + 1
    adc #0
    sta cells + 1
    lda consing
    sta LO
    lda consing + 1
    sta HI
    rts

    ; create variable
    ; -> A
op_mkvar:
    jsr fl_mkvar
    jmp next

    ; -> A
fl_mkvar:
    lda #<fl_nil
    sta X0
    lda #>fl_nil
    sta X0 + 1
    lda #0
    sta LO
    sta HI
    lda #1 << VAR_BIT
    jsr fl_cons
    lda LO
    ldy #2
    sta (LO), y
    iny 
    lda HI
    sta (LO), y
    rts

    ; runtime error
    ; X0=errmsg
error:
    lda #<errhdr
    sta X1
    lda #>errhdr
    sta X1 + 1
    jsr writestr
    lda X0
    sta X1
    lda X0 + 1
    sta X1 + 1
    jsr writestr
    lda #<errinfo1
    sta X1
    lda #>errinfo1
    sta X1 + 1
    jsr writestr
    ldy #0
    lda (GOAL), y
    sta LO
    iny
    lda (GOAL), y
    sta HI
    jsr writehex
    lda #<errinfo2
    sta X1
    lda #>errinfo2
    sta X1 + 1
    jsr writestr
    lda #13
    jsr CHROUT
    jmp exit
errhdr: .asc 13, 10, "ERROR: ", 0
errinfo1: .asc " [", 0
errinfo2: .asc "]", 13, 10, 0

    ; write string to current output device
    ; X1=string, zero terminated
writestr:
    ldy #0
lwrs0:
    lda (X1), y
    beq lwrs1
    jsr CHROUT
    iny
    jmp lwrs0
lwrs1:
    rts

    ; write P0 (string) and P1 (hex)
op_dbg:
    lda PARAM0
    sta LO
    lda PARAM0 + 1
    sta HI
    jsr fl_check_atom
    lda LO
    clc
    adc #1
    sta X1
    lda HI
    adc #0
    sta X1 + 1
    jsr writestr
    lda PARAM1
    sta LO
    lda PARAM1 + 1
    sta HI
    jsr writehex
    lda #13
    jsr CHROUT
    jmp next

    ; initialize runtime system
fl_initsys:
    ; init vars
    lda #0
    sta cells
    sta cells + 1
    sta goals
    sta goals + 1
    sta suspended
    sta suspended + 1
    sta tailcalls
    lda #2
    sta epoch
    ; init sstack
    lda #<sstack
    sta sstackp
    lda #>sstack
    sta sstackp + 1
    ; init goals
    lda #<goalbuf
    sta LO
    lda #>goalbuf
    sta HI
    lda #>MAX_GOALS
    ldx #10             ; sizeof(goal)
    jsr linkblocks
    lda LO
    sta freegoals
    lda HI
    sta freegoals + 1
    ; init heap
    lda #<HEAP
    sta LO
    lda #>HEAP
    sta HI
    lda #>MAX_CELLS
    ldx #6              ; sizeof(cell)
    jsr linkblocks
    lda LO
    sta freelist
    lda HI
    sta freelist + 1
    ; init queue
    lda #<queue
    sta qstart
    sta qend
    lda #>queue
    sta qstart + 1
    sta qend + 1
    jsr fl_mkvar
    jsr fl_addref0
    lda LO
    sta idle
    lda HI
    sta idle + 1
    rts

    ; link sequence of blocks together with ptr in firsts word
    ; x=blocksize, A=startaddr, a=multiples of 256 -> A (ptr)
linkblocks:
    sta X1 + 1          ; multiples
    lda #0
    ldy #0
    sta (LO), y         ; goal->addr
    iny
    sta (LO), y
    lda LO
    pha
    lda HI
    pha
    stx X1              ; blocksize
    lda LO
    clc
    adc X1
    sta LO
    lda HI
    adc #0
    sta HI
l100:
    ldx #0
l5:
    pla
    ldy #1
    sta (LO), y         ; goal->addr
    pla
    dey
    sta (LO), y
    lda LO
    pha
    lda HI
    pha
    lda LO
    clc
    adc X1
    sta LO
    lda HI
    adc #0
    sta HI
    dex
    bne l5
    dec X1 + 1
    bne l100
    pla 
    sta HI
    pla
    sta LO
    rts

    ; increase refcount
    ; A=val -> A
fl_addref:
    lda LO
    and #1 << IMMEDIATE_BIT
    beq ladd1
ladd2:
    rts
ladd1:
    bit HI
    bpl ladd2       ; test LOW_BIT
fl_addref0:
    ldy #0
    lda (LO), y
    cmp #COUNT_MASK
    beq ladd2
    clc
    adc #1
    sta (LO), y
    rts

    ; decrease refcount
    ; X0=val, must preserve A
fl_unref:
    lda X0
    and #1 << IMMEDIATE_BIT
    bne ladd2
    bit X0 + 1
    bpl ladd2       ; test LOW_BIT
    ldy #0
    lda (X0), y
    cmp #COUNT_MASK
    beq ladd2
    sec
    sbc #1
    sta (X0), y
    bne ladd2
    ; release cell:
    ldy #1
    lda (X0), y             ; save tag
    pha
    lda freelist + 1
    sta (X0), y
    dey
    lda freelist
    sta (X0), y
    lda X0
    sta freelist
    lda X0 + 1
    sta freelist + 1
    lda cells
    sec
    sbc #1
    sta cells
    lda cells + 1
    sbc #0
    sta cells + 1
    pla
    tax
    and #1 << INT_BIT
    beq lrel1
    rts
lrel1:
    txa
    and #1 << VAR_BIT   ; variable?
    beq l14
    ldy #2              ; unbound?
    lda (X0), y
    cmp X0
    bne l14
    iny
    lda (X0), y
    cmp X0 + 1
    bne l14
    ldy #4             ; bound to itself
    lda (X0), y
    tax
    iny
    lda (X0), y
    sta X0 + 1
    stx X0
    jmp fl_unref
l14:
    ldy #4
    lda (X0), y
    pha
    iny
    lda (X0), y
    pha
    ldy #2
    lda (X0), y
    tax
    iny
    lda (X0), y
    sta X0 + 1
    stx X0
    jsr fl_unref
    pla
    sta X0 + 1
    pla
    sta X0
    jmp fl_unref

    ; dereference value
    ; A=val -> A
fl_deref:
    lda LO
    and #1 << IMMEDIATE_BIT
    beq lder1
lder2:
    rts
lder1:
    ldy #1
    lda (LO), y
    and #1 << VAR_BIT
    beq lder2
    iny
    lda (LO), y
    cmp LO
    bne l17
    iny
    lda (LO), y
    cmp HI
    beq lder2
l17:
    ldy #2
    lda (LO), y
    tax
    iny
    lda (LO), y
    sta HI
    stx LO
    jmp fl_deref

    ; force variables
    ; A=val -> A, CF set to 1 when suspending
    ; modifies: X0
fl_force:
    lda LO
    and #1 << IMMEDIATE_BIT
    bne clrcarry
    ldy #1
    lda (LO), y
    and #1 << VAR_BIT
    beq clrcarry
    iny
    lda (LO), y
    cmp LO
    bne lfor1
    iny
    lda (LO), y
    cmp HI
    beq sspush
lfor1:
    ldy #3
    lda (LO), y
    tax
    dey
    lda (LO), y
    sta LO
    stx HI
    jmp fl_force
sspush:
    ldy #0
    lda LO
    sta (sstackp), y
    iny
    lda HI
    sta (sstackp), y
    dey
    lda (GOAL), y
    pha
    iny
    lda (GOAL), y
    tax
    pla
    ldy #2
    sta (sstackp), y
    iny
    txa
    sta (sstackp), y
    lda sstackp
    clc
    adc #4
    sta sstackp
    bne lfor2
    jmp essfull
lfor2:
    sta sstackp
setcarry:
    sec
    rts
clrcarry:
    clc
    rts

    ; allocate goal
    ; -> N
fl_opengoal:
    lda freegoals
    ora freegoals + 1
    bne l18
    lda #<ENOGOALS
    sta X0
    lda #>ENOGOALS
    sta X0 + 1
    jmp error
l18:
    lda freegoals
    sta NGOAL
    lda freegoals + 1
    sta NGOAL + 1
    ldy #0
    lda (NGOAL), y
    sta freegoals
    iny
    lda (NGOAL), y
    sta freegoals + 1
    clc
    lda goals
    adc #1
    sta goals
    lda goals + 1
    adc #0
    sta goals + 1
    ; fall through

    ; enqueue goal
    ; N=goal
enqueue:
    ldy #0
    lda NGOAL
    sta (qend), y
    iny
    lda NGOAL + 1
    sta (qend), y
    clc
    lda qend
    adc #2
    sta qend
    lda qend + 1
    adc #0
    sta qend + 1
    lda qend
    cmp qstart
    bne l19
    lda qend + 1
    cmp qstart + 1
    bne l19a
    lda #<EQUEUEFULL
    sta X0
    lda #>EQUEUEFULL
    sta X0 + 1
    jmp error
l19:
    lda qend + 1
l19a:
    cmp #>queuetop
    bne l20
    lda qend
    cmp #<queuetop
    bne l20
    lda #<queue
    sta qend
    lda #>queue
    sta qend + 1
l20:
    rts

    ; dequeue goal
    ; -> G
    ; modifies: X0
dequeue:
    ldy #0
    lda (qstart), y
    sta GOAL
    iny
    lda (qstart), y
    sta GOAL + 1
    clc
    lda qstart
    adc #2
    sta qstart
    lda qstart + 1
    adc #0
    sta qstart + 1
    lda qstart
    cmp #<queuetop
    bne l21
    lda qstart + 1
    cmp #>queuetop
    bne l21
    lda #<queue
    sta qstart
    lda #>queue
    sta qstart + 1
l21:
    rts

    ; prepare for tail call
    ; <AP>=addr -> N
    ; modifies: X1
op_tailgoal:
    lda tailcalls
    cmp timeslice
    bne l70
    jsr fl_opengoal
    jsr fetch
l70a:
    ldy #0
    lda X1
    sta (NGOAL), y      ; goal->addr
    iny
    lda X1 + 1
    sta (NGOAL), y
    jmp next
l70:
    jsr fetch
    lda GOAL
    sta NGOAL
    lda GOAL + 1
    sta NGOAL + 1
    jmp l70a

    ; invoke predicate if timeslice is not exhausted
    ; G=N
    ; modifies: X0
op_tailjump:
    lda tailcalls
    cmp timeslice
    bne l71
    lda #0
    sta tailcalls
    jmp op_dropgoal
l71:
    clc
    adc #1
    sta tailcalls
    ; fall through ...

    ; invoke predicate
    ; N
op_invoke_jump:
    ldx stackbase
    txs
    ldy #0
    lda (NGOAL), y      ; goal->addr
    sta AP
    iny
    lda (NGOAL), y
    sta AP + 1
    jmp next

    ; prepare invoking predicate
    ; G -> N, A -> addr(N)
op_invoke_goal:
    lda GOAL
    sta NGOAL
    lda GOAL + 1
    sta NGOAL + 1
    ldy #1
    lda HI
    and #$7f
    sta (NGOAL), y
    dey
    lda LO
    and #$fe
    sta (NGOAL), y
    jmp next

    ; N, A=list
op_puttailarg:
    ldy #4
    lda (LO), y
    and #1 << IMMEDIATE_BIT
    beq lpta1
    ldy #2
    lda (LO), y
    tax
    iny
    lda (LO), y
    stx LO
    sta HI
lpta1:
    jsr fl_addref
    lda LO
    ldy #8
    sta (NGOAL), y
    iny
    lda HI
    sta (NGOAL), y
    jmp next

    ; drop goal & yield
    ; G
op_dropgoal:
    lda freegoals
    ldy #0
    sta (GOAL), y       ; goal->addr
    lda freegoals + 1
    iny
    sta (GOAL), y
    lda GOAL
    sta freegoals
    lda GOAL + 1
    sta freegoals + 1
    lda goals
    sec
    sbc #1
    sta goals
    lda goals + 1
    sbc #0
    sta goals + 1
    jmp op_yield

    ; does variable have active suspension?
    ; X0: var
    ; modifies: X1, X0
active:
    ldy #4
    lda (X0), y         ; s-list
    tax
    iny
    lda (X0), y
    cpx #<fl_nil
    bne lact1
    cmp #>fl_nil
    bne lact1
    jmp clrcarry
lact1:
    stx X0
    sta X0 + 1
    ldy #2
    lda (X0), y
    sta X1
    iny
    lda (X0), y
    sta X1 + 1
    dey                 ; 2
    lda (X1), y         ; check goalptr from CAR for being 0
    iny                 ; 3
    ora (X1), y
    beq active
    jmp setcarry

    ; yield to next process
op_yield:
    jsr check_stop
    lda ticks
    clc
    adc #1
    sta ticks
    lda ticks + 1
    adc #0
    sta ticks + 1
    lda qstart
    cmp qend
    bne l22
    lda qstart + 1
    cmp qend + 1
    bne l22
    lda idle          ; queue empty, check idle var for having suspensions
    sta X0
    lda idle + 1
    sta X0 + 1
    jsr active
    bcc terminate
    lda idle
    sta X0
    lda idle + 1
    sta X0 + 1
    lda #<fl_nil
    sta LO
    lda #>fl_nil
    sta HI
    jsr fl_assign
    jsr fl_unref
    inc epoch
    jsr fl_mkvar
    jsr fl_addref0
    lda LO
    sta idle
    lda HI
    sta idle + 1
l22:
    jsr dequeue
    ; enter goal code
    ldx stackbase
    txs
    ldy #0
    lda (GOAL), y
    sta AP
    iny
    lda (GOAL), y
    sta AP + 1
    jmp next

    ; check STOP key
check_stop:
    lda STKEY
    cmp #127
    bne lcst1
    lda #<ESTOPPED
    sta X0
    lda #>ESTOPPED
    sta X0 + 1
    jmp error    
lcst1:
    rts

    ; add goal to pool
    ; X0=addr
fl_initgoal:
    jsr fl_opengoal
    lda X0
    ldy #0
    sta (NGOAL), y      ; goal->addr
    lda X0 + 1
    iny
    sta (NGOAL), y
    rts

terminate:
    lda suspended
    beq lterm1
    lda #<EDEADLOCK
    sta X0
    lda #>EDEADLOCK
    sta X0 + 1
    jmp error
lterm1:
    lda idle
    sta X0
    lda idle + 1
    sta X0 + 1
    jsr fl_unref
    lda cells
    bne eleak
    lda cells + 1
    bne eleak
exit:
    ldx stackbase
    txs
    ; enable BASIC ROM
    lda R6510
    ora #1
    sta R6510
    rts
eleak:
    lda #<ELEAK
    sta X0
    lda #>ELEAK
    sta X0 + 1
    jmp error

    ; assign value to variable
    ; A=val, X0=var 
fl_assign:
    ldy #2
    lda (X0), y
    cmp LO
    bne lass1
    iny
    lda (X0), y
    cmp HI
    bne lass1
    rts
lass1:
    ldy #2
    lda (X0), y
    cmp X0
    bne l24
    iny
    lda (X0), y
    cmp X0 + 1
    bne l24
    lda LO
    ldy #2
    sta (X0), y
    iny
    lda HI
    sta (X0), y
    jsr fl_addref
    lda X0
    sta LO
    lda X0 + 1
    sta HI
    jmp resume
l24:
    lda #<EBOUND
    sta X0
    lda #>EBOUND
    sta X0 + 1
    jmp error

    ; suspend on variables on sstack & yield
    ; G
    ; modifies: A, X0, X1
op_suspend:
    lda #<fl_nil
    sta slst
    lda #>fl_nil
    sta slst + 1
    jsr build_slst
    lda GOAL
    sta LO
    lda GOAL + 1
    sta HI
    lda slst
    sta X0
    lda slst + 1
    sta X0 + 1
    lda #1 << LIST_BIT
    jsr fl_cons           ; slst = cons(goal, slst)
    lda LO
    sta slst
    lda HI
    sta slst + 1
    jsr attach
    lda #<sstack
    sta sstackp
    lda suspended
    clc
    adc #1
    sta suspended
    lda suspended + 1
    adc #0
    sta suspended + 1
    jmp op_yield
slst: .word 0

build_slst:
    jsr for_sstack
    ; build slst: A=var, X0=addr
    lda LO
    ora #1          ; mark var
    sta LO
    lda #1 << LIST_BIT
    jsr fl_cons       ; x = cons(var', addr)
    lda slst
    sta X0
    lda slst + 1
    sta X0 + 1
    lda #1 << LIST_BIT
    jsr fl_cons       ; slst = cons(x, slst)
    lda LO
    sta slst
    lda HI
    sta slst + 1
    rts

attach:
    jsr for_sstack
    ; attach slst to all vars: A=var
    lda LO
    pha
    lda HI
    pha
    ldy #4
    lda (LO), y
    sta X0
    iny
    lda (LO), y
    sta X0 + 1
    lda slst
    sta LO
    lda slst + 1
    sta HI
    lda #1 << LIST_BIT
    jsr fl_cons       ; var->cdr = cons(slst, var->cdr)
    jsr fl_addref0
    jsr fl_unref    ; old var->cdr
    pla
    sta X0 + 1
    pla
    sta X0
    ldy #4
    lda LO
    sta (X0), y
    lda HI
    iny
    sta (X0), y
    rts

    ; resume goals on variable
    ; A=var
    ; modifies: X0, up1, up2
resume:
    lda LO
    sta rsave
    lda HI
    sta rsave + 1
    ldy #4              ; get slst
    lda (LO), y
    pha
    tax
    iny
    lda (LO), y
    pha
    ; iterate over slst of var in A:
    sta up1 + 1
    stx up1
resume1l:
    lda up1
    cmp #<fl_nil
    bne resume1
    lda up1 + 1
    cmp #>fl_nil
    bne resume1
    ; get original slst and unref:
    pla
    sta X0 + 1
    pla
    sta X0
    jsr fl_unref
    ; store nil in slst of var:
    lda rsave
    sta X0
    lda rsave + 1
    sta X0 + 1
    lda #<fl_nil
    ldy #4
    sta (X0), y
    lda #>fl_nil
    iny
    sta (X0), y
    rts
resume1:
    ; check slst-entry [Goal, Vars/Addrs, ...] of var
    ldy #2
    lda (up1), y
    sta LO
    iny
    lda (up1), y
    sta HI
    dey
    ; goal not marked (0)?
    lda (LO), y
    iny
    ora (LO), y
    bne lres1
resume1c:
    ; get cdr of slst and continue loop:
    ldy #4
    lda (up1), y
    tax
    iny
    lda (up1), y
    sta up1 + 1
    stx up1
    jmp resume1l
lres1:
    ; enqueue goal from slst entry:
    ldy #2
    lda (LO), y
    sta NGOAL
    iny
    lda (LO), y
    sta NGOAL + 1
    jsr enqueue
    lda suspended
    sec
    sbc #1
    sta suspended
    lda suspended + 1
    sbc #0
    sta suspended + 1
    lda #0
    ldy #2
    sta (LO), y      ; clear goal
    iny
    sta (LO), y
    ldy #4
    lda (LO), y
    sta up2
    iny
    lda (LO), y
    sta up2 + 1    
resume2l:
    ; iterate over [Var, Addr] pairs:
    lda up2
    cmp #<fl_nil
    bne lrs2
    lda up2 + 1
    cmp #>fl_nil
    bne lrs2
    jmp resume1c
lrs2:
    ; find variable/addr pair in slst
    ldy #2
    lda (up2), y
    sta LO
    iny
    lda (up2), y
    sta HI
    dey                 ; 2
    lda (LO), y
    and #$fe            ; unmark var
    sta X0
    iny                 ; 3
    lda (LO), y
    sta X0 + 1
    lda rsave
    cmp X0
    beq lres2
resume2c:
    ldy #4
    lda (up2), y
    tax
    iny
    lda (up2), y
    sta up2 + 1
    stx up2
    jmp resume2l
lres2:
    lda rsave + 1
    cmp X0 + 1
    bne resume2c
    ; set addr of freshly enqueued goal:
    ldy #4
    lda (LO), y
    tax
    iny
    lda (LO), y
    ldy #1
    sta (NGOAL), y      ; goal->addr
    dey
    txa
    sta (NGOAL), y
    jmp resume1c
rsave: .word 0

    ; match failed
op_fail:
    lda #<EFAIL
    sta X0
    lda #>EFAIL
    sta X0 + 1
    jmp error

    ; abort with error
    ; P0=atom
op_abort:
    lda PARAM0
    sta LO
    lda PARAM0 + 1
    sta HI
    jsr fl_check_atom
    lda LO
    clc
    adc #1
    sta X0
    lda HI
    adc #0
    sta X0 + 1
    jmp error

    ; unify with result
    ; P0=val1, P1=val2 -> A (true or false)
op_unify:
    jsr setup
    jsr unify
    bcs luni1
    jmp op_true
luni1:
    jmp op_false
    
    ; unify with check
    ; P0=val1, P1=val2
op_unify_checked:
    jsr setup
    jsr unify
    bcs eunify
    jmp next
eunify:
    lda #<EUNIFY
    sta X0
    lda #>EUNIFY
    sta X0 + 1
    jmp error

    ; unify result with check
    ; A=val1, X0=val2
fl_unify_result:
    lda LO
    sta up1
    lda HI
    sta up1 + 1
    lda X0
    sta up2
    lda X0 + 1
    sta up2 + 1
    jsr unify
    bcs eunify
    rts

upderef:
    lda up1
    sta LO
    lda up1 + 1
    sta HI
    jsr fl_deref
    lda LO
    sta up1
    lda HI
    sta up1 + 1
    lda up2
    sta LO
    lda up2 + 1
    sta HI
    jsr fl_deref
    lda LO
    sta up2
    lda HI
    sta up2 + 1
    rts

    ; unify values
    ; up1=val1, up2=val2 -> Carry (1 is mismatch)
unify:
    jsr upderef
    lda up1
    cmp LO
    bne l80
    lda up1 + 1
    cmp HI
    bne l80
    jmp clrcarry
l80:
    ; X != Y
    lda up1
    and #1 << IMMEDIATE_BIT
    bne l82
    ldy #1
    lda (up1), y
    and #1 << VAR_BIT
    beq l82
    lda up1
    sta X0
    lda up1 + 1
    sta X0 + 1
    lda up2
    sta LO
    lda up2 + 1
    sta HI
    jsr fl_assign
    jmp clrcarry
l82:
    ; X is not a variable
    lda up2
    and #1 << IMMEDIATE_BIT
    bne l83b
    ; Y is not immediate
    ldy #1
    lda (up2), y
    and #1 << VAR_BIT
    beq l83
    lda up2
    sta X0
    lda up2 + 1
    sta X0 + 1
    lda up1
    sta LO
    lda up1 + 1
    sta HI
    jsr fl_assign
    jmp clrcarry
l83:
    ; Y is not a variable
    lda up1
    and #1 << IMMEDIATE_BIT
    bne l83b
    ; X is not immediate
    ldy #1
    ; compare tags:
    lda (up2), y
    cmp (up1), y
    beq l83a
l83b:
    jmp setcarry
l83a:
    ; tags are equal
    ldy #4
    lda (up1), y
    pha
    iny
    lda (up1), y
    pha
    ldy #2
    lda (up1), y
    tax
    iny
    lda (up1), y
    sta up1 + 1
    stx up1
    ldy #4
    lda (up2), y
    pha
    iny
    lda (up2), y
    pha
    ldy #2
    lda (up2), y
    tax
    iny
    lda (up2), y
    sta up2 + 1
    stx up2
    jsr unify
    bcc l84
    pla
    pla
    pla
    pla
    rts
l84:
    pla
    sta up1 + 1
    pla
    sta up1
    pla
    sta up2 + 1
    pla
    sta up2
    jmp unify

    ; match values
    ; G, P0=val1, P1=val2 -> A (true, false or nil)
op_match:
    jsr setup
    jsr match
    bcs op_nil
    beq lmat2
    jmp op_false
lmat2:
    jmp op_true
op_nil:
    lda #<fl_nil
    sta LO
    lda #>fl_nil
    sta HI
    jmp next

    ; match values
    ; G, up1=val1, up2=val2 -> Z + C (Z set if match, C set on suspend)
match:
    tsx
    stx ssave
match_rec:
    jsr upderef
    lda up1
    cmp up2
    bne l85
    lda up1 + 1
    cmp up2 + 1
    bne l85
    jmp mtrue1
l85:
    ; not eq
    lda up1
    and #1 << IMMEDIATE_BIT
    bne l86
    ldy #1
    lda (up1), y
    and #1 << VAR_BIT
    beq l86
lmr1:
    jsr save_and_suspend
    ldx ssave
    txs
    sec
    rts
l86:
    ; X is not a variable
    lda up2
    and #1 << IMMEDIATE_BIT
    bne mfalse
    ldy #1
    lda (up2), y
    and #1 << VAR_BIT
    beq l87
    lda up2
    sta up1
    lda up2 + 1
    sta up1 + 1
    jmp lmr1
l87:
    ; Y is not a variable
    lda up1
    and #1 << IMMEDIATE_BIT
    bne mfalse
    ; X and Y are not immediate and not variables
    ldy #1
    lda (up2), y
    cmp (up1), y
    bne mfalse
    ; tags are equal
    ldy #4
    lda (up1), y
    pha
    iny
    lda (up1), y
    pha
    ldy #2
    lda (up1), y
    tax
    iny
    lda (up1), y
    sta up1 + 1
    stx up1
    ldy #4
    lda (up2), y
    pha
    iny
    lda (up2), y
    pha
    ldy #2
    lda (up2), y
    tax
    iny
    lda (up2), y
    sta up2 + 1
    stx up2
    jsr match_rec
    beq l87c
    ; mismatch, pop stack and return
    bcc mfalse
    lda up2
    sta up1
    lda up2 + 1
    sta up1 + 1
    jmp lmr1
l87c:
    pla
    sta up2 + 1
    pla
    sta up2
    pla 
    sta up1 + 1
    pla
    sta up1
    jmp match_rec
ssave: .byte 0
mtrue:
    ldx ssave
    txs
mtrue1:
    and #0           ; set Z, clear C
    clc
    rts
mfalse:
    ldx ssave
    txs
    ora #1            ; clear Z, clear C
    clc
    rts

save_and_suspend: ; (up1 is var)
    ldy #0
    lda up1
    sta (sstackp), y
    iny
    lda up1 + 1
    sta (sstackp), y
    iny
    lda GOAL
    sta (sstackp), y
    iny
    lda GOAL + 1
    sta (sstackp), y
    lda sstackp
    clc
    adc #4
    sta sstackp
    bne lss1a
essfull:
    lda #<ESSFULL
    sta X0
    lda #>ESSFULL
    sta X0 + 1
    jmp error
lss1a:
    rts

    ; compare for equality
    ; P0, P1 -> Carry (set if not equal)
fl_eq:
    lda PARAM0
    cmp PARAM1
    bne leq1
    lda PARAM0 + 1
    cmp PARAM1 + 1
    bne leq1
    jmp clrcarry
leq1:
    jmp setcarry

    ; A=val -> A (int)
op_intval:
    jsr fl_getint
    jmp next

    ; A=int -> A
op_to_int:
    jsr fl_mkint
    jmp next

    ; arithmetic
    ; (SP) + A -> A (int)
op_add:
    pla
    tax
    pla
    clc
    adc LO
    sta LO
    txa
    adc HI
    sta HI
    jmp next

    ; (SP) - A -> A (int)
op_sub:
    pla
    tax
    pla
    sec
    sbc LO
    sta LO
    txa
    sbc HI
    sta HI
    jmp next

    ; (SP) * A -> A (int)
op_mul:
    lda LO
    sta X1
    lda HI
    sta X1 + 1
    pla
    sta X0 + 1
    pla
    sta X0
    ; http://map.grauw.nl/articles/mult_div_shifts.php
    ; A = X1 * X0
    ldy #16
Mult16_Loop:
    clc
    rol LO
    rol HI
    clc
    rol X1
    rol X1 + 1
    bcc Mult16_NoAdd
    lda X0
    clc
    adc LO
    sta LO
    lda X0 + 1
    adc HI
    sta HI
Mult16_NoAdd:
    dey
    bne Mult16_Loop
    jmp next

    ; (SP) / A -> A (int)
op_div:
    lda LO
    sta up2
    lda HI
    sta up2 + 1
    pla
    sta up1 + 1
    pla
    sta up1
    lda #0
    sta divsign
    jsr divmod
    lda X0
    sta LO
    lda X0 + 1
    sta HI
    lda divsign
    beq ldivn1
    jsr negate
ldivn1:
    jmp next

    ; (SP) % A -> A (int)
op_mod:
    lda LO
    sta up2
    lda HI
    sta up2 + 1
    pla
    sta up1 + 1
    pla
    sta up1
    jsr divmod
    jmp next

    ; up1 / up2 -> X0 + A
divsign: .byte 0
divmod:
    lda up1
    sta LO
    lda up1 + 1
    sta HI
    bpl ldiv2
    jsr negate
ldiv2:
    lda LO
    sta X0
    lda HI
    sta X0 + 1
    lda up2
    sta LO
    lda up2 + 1
    sta HI
    bpl ldiv3
    jsr negate
ldiv3:
    lda LO
    sta X1
    lda HI
    sta X1 + 1
    lda up1 + 1
    eor up2 + 1
    bpl ldiv1
    lda #1
    sta divsign
ldiv1:
    lda X1
    bne l120
    lda X1 + 1
    bne l120
    lda #<EDIV0
    sta X0
    lda #>EDIV0
    sta X0 + 1
    jmp error
l120:
    ; https://codebase64.org/doku.php?id=base:16bit_division_16-bit_result
    ; X0 + A = X0 / X1
divide:
    lda #0          ;preset remainder to 0
    sta LO          ;remainder
    sta HI          ;remainder+1
    ldx #16         ;repeat for each bit: ... 
divloop:
    asl X0          ;dividend lb & hb*2, msb -> Carry
    rol X0+1        ; dividend + 1
    rol LO          ;remainder lb & hb * 2 + msb from carry
    rol HI          ;remainder+1
    lda LO          ;remainder
    sec
    sbc X1          ;substract divisor to see if it fits in
    tay             ;lb result -> Y, for we may need it later
    lda HI          ;remainder+1
    sbc X1 + 1      ;divisor+1
    bcc divskip        ;if carry=0 then divisor didn't fit in yet
    sta HI          ;else save substraction result as new remainder,
    sty LO          ;remainder
    inc X0          ;and INCrement result cause divisor fit in 1 times
divskip:
    dex
    bne divloop     
    rts

    ; integer square root
    ; A (int) -> A (int)
op_sqrt:
    ; http://map.grauw.nl/sources/external/z80bits.html#3.2
    lda #$ff
    tax
    sta X0
    sta X0 + 1
Sqr16_Loop:
    clc
    lda LO
    adc X0
    sta LO
    lda HI
    adc X0 + 1
    sta HI
    php
    inx
    dec X0
    sec
    lda X0
    sbc #1
    sta X0
    lda X0 + 1
    sbc #0
    sta X0 + 1
    plp
    bcs Sqr16_Loop
    lda #0
    sta HI
    txa
    sta LO
    jmp next

    ; binary and
    ; (SP) & A (int) -> A (int)
op_and:
    pla
    and HI
    sta HI
    pla
    and LO
    sta LO
    jmp next

    ; binary or
    ; (SP) | A (int) -> A (int)
op_or:
    pla
    ora HI
    sta HI
    pla
    ora LO
    sta LO
    jmp next

    ; binary xor
    ; (SP) ^ A (int) -> A (int)
op_xor:
    pla
    eor HI
    sta HI
    pla
    eor LO
    sta LO
    jmp next

    ; binary shift left
    ; (SP) << A (int) -> A (int)
op_shl:
    ldx LO
    pla
    sta HI
    pla
    sta LO
    txa
    bne lshl1
    jmp next
lshl1:
    asl LO
    rol HI
    dex
    bne lshl1
    jmp next

    ; binary shift right
    ; (SP) >> A (int) -> A (int)
op_shr:
    ldx LO
    pla
    sta HI
    pla
    sta LO
    txa
    bne lshr1
    jmp next
lshr1:
    clc
    bit HI
    bpl lshr2
    sec
lshr2:
    ror HI
    ror LO
    dex
    bne lshr1
    jmp next

    ; match atom or small int
    ; A=val, <AP>=atomic + goaladdr + jumpaddr
op_isval:
    jsr fetch
    lda X1
    cmp LO
    bne lisv1
    lda X1 + 1
    cmp HI
    bne lisv1
    ; fall through...
skip2:
    jsr skip
    jmp op_skip
lisv1:
    jmp trynext

    ; match boxed integer
    ; A=val, <AP>=int + goaladdr + jumpaddr
op_isint:
    jsr fetch
    lda LO
    and #1 << IMMEDIATE_BIT
    bne lisi1
    ldy #1
    lda (LO), y
    and #1 << INT_BIT
    beq lisi1
    iny
    lda (LO), y
    cmp X1
    bne lisi1
    iny
    lda (LO), y
    cmp X1 + 1
    beq skip2
lisi1:
    jmp trynext

    ; check value for being false and jump if it isn't
    ; A=val, <AP>=jumpaddr
op_isfalse:
    lda LO
    cmp #<fl_false
    bne lisf1
    lda HI
    cmp #>fl_false
    beq op_skip
lisf1:
    jmp op_jump

op_skip:
    jsr skip
    jmp next

    ; check value for being true and jump if it isn't
    ; A=val, <AP>(jumpaddr)
op_istrue:
    lda LO
    cmp #<fl_true
    bne list1a
    lda HI
    cmp #>fl_true
    beq op_skip
list1a:
    jmp op_jump

    ; check value for being atomic (small int or atom)
    ; A=val, <AP>=addr
op_isatomic:
    lda LO
    and #1 << IMMEDIATE_BIT
    beq list1a
    jmp op_skip

    ; check value for being list
    ; A=val -> C (set if not list)
fl_islist_t:
    lda LO
    and #1 << IMMEDIATE_BIT
    beq lislt1
lislt2:
    jmp setcarry
lislt1:
    ldy #1
    clc
    lda (LO), y
    and #1 << LIST_BIT
    beq lislt2
    rts

    ; check value for being list
    ; A=val, <AP>=jumpaddr
op_islist_t:
    jsr fl_islist_t
    bcs lislt1a
    jmp op_skip
lislt1a:
    jmp op_jump

    ; check value for being list and deconstruct
    ; A=val, <AP>=jumpaddr
op_islist_n:
    jsr fl_islist_t
    bcs lislt1a
    jsr skip
    jmp op_decons

    ; check value for being list and deconstruct
    ; A=val, <AP>=goaladdr + jumpaddr -> A(car), cdr on stack
op_islist:
    jsr fl_islist_t
    bcc lisl1
    jmp trynext
lisl1:
    jsr skip
    jsr skip
    ; fall through ...

    ; deconstruct cell
    ; A=val -> A(car), cdr on stack
op_decons:
    ldy #4
    lda (LO), y
    pha
    iny
    lda (LO), y
    pha
    ldy #2
    lda (LO), y
    tax
    iny
    lda (LO), y
    sta HI
    stx LO
    jmp next

    ; check value for being tuple and deconstruct
    ; A=val, <AP>=arity + goaladdr + jumpaddr -> A (head), arglist on stack
op_istuple:
    jsr fetch
    lda LO
    and #1 << IMMEDIATE_BIT
    bne list2
    ldy #1
    lda (LO), y
    and #1 << TUPLE_BIT
    beq list2
    lda (LO), y
    and #ARITY_MASK >> 8
    cmp X1
    beq list1
list2:
    jmp trynext
list1:
    jsr skip
    jsr skip
    jmp op_decons

    ; check for numeric or immediate equality
    ; P0=val1, P1=val2 -> A
op_equal:
    lda PARAM0
    cmp PARAM1
    bne leql1
    lda PARAM0 + 1
    cmp PARAM1 + 1
    beq leql3
leql1:
    lda PARAM0
    and #1 << IMMEDIATE_BIT
    bne leql2
    lda PARAM1
    and #1 << IMMEDIATE_BIT
    bne leql2
    ldy #1
    lda (PARAM0), y
    cmp (PARAM1), y
    bne leql2
    iny
    lda (PARAM0), y
    cmp (PARAM1), y
    bne leql2
    iny
    lda (PARAM0), y
    cmp (PARAM1), y
    bne leql2
leql3:
    jmp op_true
leql2:
    jmp op_false

    ; check for numerically less-than
    ; P0=val1, P1=val2 -> A
op_less:
    lda PARAM0
    sta LO
    lda PARAM0 + 1
    sta HI
    jsr fl_getint
    lda LO
    sta up1
    lda HI
    sta up1 + 1
    lda PARAM1
    sta LO
    lda PARAM1 + 1
    sta HI
    jsr fl_getint
    lda LO
    sta up2
    lda HI
    sta up2 + 1
    jsr less
    bcc lgrt1
    jmp op_false
 
    ; check for numerically greater-than
    ; P0=val1, P1=val2 -> A
op_greater:
    lda PARAM0
    sta LO
    lda PARAM0 + 1
    sta HI
    jsr fl_getint
    lda LO
    sta up2
    lda HI
    sta up2 + 1
    lda PARAM1
    sta LO
    lda PARAM1 + 1
    sta HI
    jsr fl_getint
    lda LO
    sta up1
    lda HI
    sta up1 + 1
    jsr less
    bcc lgrt1
    jmp op_false
lgrt1:
    jmp op_true

    ; from "Programming the Z80", Rodney Zacks, p.531
    ; up1=int1, up2=int2 -> C (set if NOT less)
less:
    lda up1 + 1
    bmi lles1
    lda up2 + 1
    bpl lles2
    sec
lles4:
    rts
lles2:
    lda up1 + 1
    cmp up2 + 1
    bne lles4
    lda up1
    cmp up2
    rts
lles1:
    bit up2 + 1
    bmi lles5
    clc
    rts
lles5:
    lda up1 + 1
    cmp up2 + 1
    bne lles4
    lda up1
    cmp up2
    rts

    ; update goal-ptr and continue at new address
    ; <AP>=addr + jumpaddr, G
    ; modifies: X1
trynext:
    jsr fetch
    lda X1
    ldy #0
    sta (GOAL), y       ; goal->addr
    lda X1 + 1
    iny
    sta (GOAL), y
    jmp op_jump

    ; type check for integer
    ; A=val
fl_check_integer:
    lda LO
    and #1 << IMMEDIATE_BIT
    beq lchki1
    bit HI            ; test LOW_BIT
    bpl lchki1
lchki2:
    rts
lchki1:
    ldy #1
    lda (LO), y
    and #1 << INT_BIT
    bne lchki2
    lda #<ENOINT
    sta X0
    lda #>ENOINT
    sta X0 + 1
    jmp error

    ; type check for string
    ; A=val
fl_check_atom:
    lda LO
    and #1 << IMMEDIATE_BIT
    beq lchka1
    bit HI
    bmi lchka1      ; test LOW_BIT
    rts
lchka1:
    lda #<ENOSTRING
    sta X0
    lda #>ENOSTRING
    sta X0 + 1
    jmp error

    ; type check for cell
    ; A=val
fl_check_cell:
    lda LO
    and #1 << IMMEDIATE_BIT
    bne lchkc1
    rts
lchkc1:
    lda #<ENOCELL
    sta X0
    lda #>ENOCELL
    sta X0 + 1
    jmp error

    ; type check for tuple
    ; A=val
fl_check_tuple:
    lda LO
    and #1 << IMMEDIATE_BIT
    bne lchkt1
    ldy #1
    lda (LO), y
    and #1 << TUPLE_BIT
    beq lchkt1
    rts
lchkt1:
    lda #<ENOTUPLE
    sta X0
    lda #>ENOTUPLE
    sta X0 + 1
    jmp error

    ; type check for list
    ; A=val
fl_check_list:
    lda LO
    and #1 << IMMEDIATE_BIT
    bne lchkl1
    ldy #1
    lda (LO), y
    and #1 << LIST_BIT
    beq lchkl1
    rts
lchkl1:
    lda #<ENOLIST
    sta X0
    lda #>ENOLIST
    sta X0 + 1
    jmp error

    ; type check for var
    ; A=val
fl_check_var:
    lda LO
    and #1 << IMMEDIATE_BIT
    bne lchkv1
    ldy #1
    lda (LO), y
    and #1 << VAR_BIT
    beq lchkv1
    rts
lchkv1:
    lda #<ENOVAR
    sta X0
    lda #>ENOVAR
    sta X0 + 1
    jmp error

    ; test if tuple
    ; A=val, <AP>=addr 
op_istuple_t:
    lda LO
    and #1 << IMMEDIATE_BIT
    bne listt1
    ldy #1
    lda (LO), y
    and #1 << TUPLE_BIT
    beq listt1
    jmp op_skip
listt1:
    jmp op_jump

    ; test if int
    ; P0=val -> A (bool)
op_test_number:
    lda PARAM0
    and #1 << IMMEDIATE_BIT
    bne lti1
    ldy #1
    lda (PARAM0), y
    and #1 << INT_BIT
    bne lti2
    jmp op_false
lti1:
    bit PARAM0 + 1      ; test LOW_BIT
    bmi lti2
lti3:
    jmp op_false
lti2:
    jmp op_true

    ; test if atom
    ; P0=val -> A (bool)
op_test_atom:
    lda PARAM0
    and #1 << IMMEDIATE_BIT
    beq lti3
    bit PARAM0 + 1      ; test LOW_BIT
    bmi lti3
    jmp op_true

    ; test if var
    ; P0=val -> A (bool)
op_test_var:
    lda PARAM0
    and #1 << IMMEDIATE_BIT
    bne lti3
    ldy #1
    lda (PARAM0), y
    and #1 << VAR_BIT
    bne lti2
    jmp op_false

    ; test if tuple
    ; P0=val -> A (bool)
op_test_tuple:
    lda PARAM0
    and #1 << IMMEDIATE_BIT
    bne lti3
    ldy #1
    lda (PARAM0), y
    and #1 << TUPLE_BIT
    bne lti2
    jmp op_false

    ; test if list
    ; P0=val -> A (bool)
op_test_list:
    lda PARAM0
    and #1 << IMMEDIATE_BIT
    bne lti3
    ldy #1
    lda (PARAM0), y
    and #1 << LIST_BIT
    bne lti2
    jmp op_false

    ; absolute value
    ; A (int) -> A (int)
op_abs:
    lda HI
    bpl labs1
    jsr negate
labs1:
    jmp next

    ; sign of integer
    ; A=val (int) -> A (int)
op_sign:
    lda LO
    ora HI
    bne lsign1
    lda #0
    sta LO
    sta HI
    jmp next
lsign1:
    lda HI
    bpl lsign2
    lda #$ff
    sta LO
    sta HI
    jmp next
lsign2:
    lda #1
    sta LO
    lda #0
    sta HI
    jmp next

setup:
    lda PARAM0
    sta up1
    lda PARAM0 + 1
    sta up1 + 1
    lda PARAM1
    sta up2
    lda PARAM1 + 1
    sta up2 + 1
    rts

    ; ordered compare
    ; P0=val1, P1=val2 -> A (bool)
op_below:
    jsr setup
    jsr ordering
    bcs ofail
    bmi labo2
    jmp op_false
ofail:
    jmp op_nil

    ; P0=val1, P1=val2 -> A (bool)
op_above:
    jsr setup
    jsr ordering
    bcs ofail
    bmi labo1
    beq labo1
labo2:
    jmp op_true
labo1:
    jmp op_false

    ; access idle var
    ; <AP>=cache -> A (bool)
op_idle:
    jsr fetcha
    ldy #0
    lda (LO), y
    tax
    inx
    cpx epoch
    bne lidl1
    lda #0
    sta (LO), y
    jmp op_true
lidl1:
    lda epoch
    sta (LO), y
    lda idle
    sta LO
    lda idle + 1
    sta HI
    jsr sspush
    jmp op_false

    ; maximum value
    ; max((SP), A) -> A (int)
op_max:
    pla
    sta up1 + 1
    pla
    sta up1
    lda LO
    sta up2
    lda HI
    sta up2 + 1
    jsr less
    bcs lmax1
    jmp next
lmax1:
    lda up1
    sta LO
    lda up1 + 1
    sta HI
    jmp next

    ; minimum value
    ; min((SP), A) -> A (int)
op_min:
    pla
    sta up1 + 1
    pla
    sta up1
    lda LO
    sta up2
    lda HI
    sta up2 + 1
    jsr less
    bcc lmin1
    jmp next
lmin1:
    lda up1
    sta LO
    lda up1 + 1
    sta HI
    jmp next

    ; allocate list cell
    ; A=cdr, (sp + 2)=car -> A
op_mklist:
    lda LO
    sta X0
    lda HI
    sta X0 + 1
    pla
    sta HI
    pla
    sta LO
    lda #1 << LIST_BIT
    jsr fl_cons
    jmp next

    ; allocate tuple
    ; A=args, <AP>=atom + arity -> A
op_mktuple:
    lda LO
    sta X0
    lda HI
    sta X0 + 1
    jsr fetcha
    jsr fetch
    lda X1
    ora #1 << TUPLE_BIT
    jsr fl_cons
    jmp next

    ; mutate list into general tuple
    ; A=list, a=arity -> A
fl_mktuple0:
    pha
    and #$f8
    beq lmkt1
    lda #<EARITY
    sta X0
    lda #>EARITY
    sta X0 + 1
    jmp error
lmkt1:
    pla
    ldy #1
    and #ARITY_MASK >> 8
    ora #1 << TUPLE_BIT
    sta (LO), y
    rts

op_mktuple0:
    jsr fetch
    lda X1
    jsr fl_mktuple0
    jmp next

    ; threaded interpreter
next:
    jsr fetch
    jmp (X1)

    ; -> X1
    ; modifies: AP
fetch:
    ldy #0
    lda (AP), y
    sta X1
    iny
    lda (AP), y
    sta X1 + 1
    lda AP
    clc
    adc #2
    sta AP
    lda AP + 1
    adc #0
    sta AP + 1
    rts

    ; -> A
    ; modifies: AP
fetcha:
    jsr fetch
    lda X1
    sta LO
    lda X1 + 1
    sta HI
    rts

skip:
    lda AP
    clc
    adc #2
    sta AP
    lda AP + 1
    adc #0
    sta AP + 1
    rts

    ; sp: points to threaded code
fl_exec:
    pla
    sta AP
    inc AP
    pla
    sta AP + 1
    jmp next

    ; unconditional addref, expects block
    ; A=val
op_addref:
    ldy #0
    lda (LO), y
    clc
    adc #1
    sta (LO), y
    jmp next

    ; A=val
op_addrefx:
    jsr fl_addref
    jmp next

    ; P0=val, A=var
op_assign:
    lda LO
    sta X0
    lda HI
    sta X0 + 1
    lda PARAM0
    sta LO
    lda PARAM0 + 1
    sta HI
    jsr fl_assign
    jmp next

    ; test if atomic
    ; P0=val -> A (bool)
op_atomic:
    lda PARAM0
    and #1 << IMMEDIATE_BIT
    bne op_true
    ldy #1
    lda (PARAM0), y
    and #1 << INT_BIT
    bne op_true
    ; fall through

    ; -> A
op_false:
    lda #<fl_false
    sta LO
    lda #>fl_false
    sta HI
    jmp next

    ; -> A
op_true:
    lda #<fl_true
    sta LO
    lda #>fl_true
    sta HI
    jmp next

    ; <AP>=addr -> A
peek1:
    jsr fetch
    ldy #0
    lda (X1), y
    sta LO
    iny
    lda (X1), y
    sta HI
    rts

    ; <AP>=addr -> A
op_peek:
    jsr peek1
    jmp next

    ; <AP>=addr -> A
op_peek_deref:
    jsr peek1
    jmp op_deref

    ; A, <AP>=addr
op_poke:
    jsr fetch
    lda LO
    ldy #0
    sta (X1), y
    lda HI
    iny
    sta (X1), y
    jmp next

    ; A, <AP>=addr
op_store:
    jsr fl_addref
    jmp op_poke

    ; <AP>=addr, mkvar -> A
op_storevar:
    jsr fl_mkvar
    jmp op_store

    ; A=val
op_check_atom:
    jsr fl_check_atom
    jmp next
op_check_cell:
    jsr fl_check_cell
    jmp next
op_check_integer:
    jsr fl_check_integer
    jmp next
op_check_list:
    jsr fl_check_list
    jmp next
op_check_tuple:
    jsr fl_check_tuple
    jmp next
op_check_var:
    jsr fl_check_var
    jmp next

op_intval_checked:
    jsr fl_check_integer
    jmp op_intval

op_tryelse:
    ldx stackbase
    txs
    jmp next

    ; <AP> -> A
op_lit:
    jsr fetcha
    jmp next

    ; A -> A
op_deref:
    jsr fl_deref
    jmp next

    ; A -> A, <AP>=jumpaddr
op_force:
    jsr fl_force
    bcs op_jump
    jmp op_skip

    ; <AP>=jumpaddr
op_forcearg:
    ldy #2
    lda (GOAL), y
    sta LO
    iny
    lda (GOAL), y
    sta HI
    jsr fl_force
    bcs op_jump
    ldy #2
    lda LO
    cmp (GOAL), y
    bne lfoa1
    iny
    lda HI
    cmp (GOAL), y
    bne lfoa1
    jmp op_skip
lfoa1:
    jsr fl_addref
    ldy #2
    lda (GOAL), y
    sta X0
    iny
    lda (GOAL), y
    sta X0 + 1
    jsr fl_unref
    ldy #2
    lda LO
    sta (GOAL), y
    iny
    lda HI
    sta (GOAL), y
    jmp op_skip
    
    ; <AP>=addr
op_jump:
    jsr fetch
    lda X1
    sta AP
    lda X1 + 1
    sta AP + 1
    jmp next

    ; <AP>=addr
op_fcall:
    jsr fetch
    lda #>next
    pha
    lda #<next - 1
    pha
    jmp (X1)
    
    ; -> A
op_getarg1:
    ldy #2
lget1:
    lda (GOAL), y
    sta LO
    iny
    lda (GOAL), y
    sta HI
    jmp next
op_getarg2:
    ldy #4
    jmp lget1
op_getarg3:
    ldy #6
    jmp lget1
op_getarg4:
    ldy #8
    jmp lget1

    ; A
op_putarg1:
    ldy #2
lput1:
    lda LO
    sta (NGOAL), y
    iny
    lda HI
    sta (NGOAL), y
    jmp next
op_putarg2:
    ldy #4
    jmp lput1
op_putarg3:
    ldy #6
    jmp lput1
op_putarg4:
    ldy #8
    jmp lput1

    ; A
op_storearg1:
    ldy #2
lsto1:
    jsr fl_addref
    lda LO
    sta (NGOAL), y
    iny
    lda HI
    sta (NGOAL), y
    jmp next
op_storearg2:
    ldy #4
    jmp lsto1
op_storearg3:
    ldy #6
    jmp lsto1
op_storearg4:
    ldy #8
    jmp lsto1

    ; A
op_putparam1:
    lda LO
    sta PARAM0
    lda HI
    sta PARAM0 + 1
    jmp next
op_putparam2:
    lda LO
    sta PARAM1
    lda HI
    sta PARAM1 + 1
    jmp next
op_putparam3:
    lda LO
    sta PARAM2
    lda HI
    sta PARAM2 + 1
    jmp next
op_putparam4:
    lda LO
    sta PARAM3
    lda HI
    sta PARAM3 + 1
    jmp next

    ; <AP>=addr
op_not_suspending:
    lda sstackp
    beq los1
    jmp op_skip
op_suspending:
    lda sstackp
    bne los1
    jmp op_skip
los1:
    jmp op_jump

    ; P0, P1, <AP>=jumpaddr
op_guard:
    jsr fl_eq
    bcc lgrd1
    jmp op_jump
lgrd1:
    jmp op_skip

    ; <AP>=addr
op_opengoal:
    jsr fl_opengoal
    jsr fetch
    lda X1
    ldy #0
    sta (NGOAL), y      ; goal->addr
    iny
    lda X1 + 1
    sta (NGOAL), y
    jmp next

    ; -> A
op_pop:
    pla
    sta HI
    pla
    sta LO
    jmp next

    ; A
op_push:
    lda LO
    pha
    lda HI
    pha
    jmp next

op_reset:
    lda #<sstack
    sta sstackp
    jmp next

    ; <AP>=addr
op_drop:
    jsr fetch
    ldy #0
    lda (X1), y
    sta X0
    iny
    lda (X1), y
    sta X0 + 1
    jsr fl_unref
    jmp next

op_droparg1:
    ldy #2
ldropa1:
    lda (GOAL), y
    sta X0
    iny
    lda (GOAL), y
    sta X0 + 1
    jsr fl_unref
    jmp next
op_droparg2:
    ldy #4
    jmp ldropa1
op_droparg3:
    ldy #6
    jmp ldropa1
op_droparg4:
    ldy #8
    jmp ldropa1

    ; auxilliary routines

    ; ordering of types 
    ; A=val -> a (order code)
type_order:
    lda LO
    and #1 << IMMEDIATE_BIT
    beq lord1
    bit HI
    bpl lord2       ; test LOW_BIT
    lda #1     ; int
    rts
lord2:
    lda #2     ; str
    rts
lord1:
    ldy #1
    lda (LO), y
    and #1 << TUPLE_BIT
    beq lord3
    lda #4     ; tup
    rts
lord3:
    lda (LO), y
    and #1 << VAR_BIT
    beq lord4
    lda #0     ; var
    rts
lord4:
    lda (LO), y
    and #1 << INT_BIT
    beq lord5
    lda #1     ; int
    rts
lord5:
    lda #3     ; list
    rts

    ; ordering of values
    ; up1=val1, up2=val2 -> Z, S (set if P0 == P1, P0 < P1)
    ; modifies: A, X0, X1
ordering:
    tsx
    stx ssave
order_rec:
    lda up1
    sta LO
    lda up1 + 1
    sta HI
    jsr type_order
    bne lorc1
lorc2:
    jsr save_and_suspend
    sec
    rts
lorc1:
    sta X0
    lda up2
    sta LO
    lda up2 + 1
    sta HI
    jsr type_order
    bne lorc3
    lda up2
    sta up1
    lda up2 + 1
    sta up1
    jmp lorc2
lorc3:
    sta X0 + 1
    lda X0
    sec
    sbc X0 + 1
    beq loo1
    clc
    rts
loo1:
    lda X0
    asl                 ; typecode * 3
    clc
    adc X0
    tax
    lda #<cmptable
    sta X0
    lda #>cmptable
    sta X0 + 1
    txa
    adc X0
    sta X0
    lda X0 + 1
    adc #0
    sta X0 + 1
    jmp (X0)
cmptable:
    jmp cmpptr
    jmp cmpint
    jmp cmpstr
    jmp cmplst
    jmp cmptup
cmpint:
    lda up1
    sta LO
    lda up1 + 1
    sta HI
    jsr fl_getint
    lda LO
    sta X0
    lda HI
    sta X0 + 1
    lda up2
    sta LO
    lda up2 + 1
    sta HI
    jsr fl_getint
    lda X0 + 1
    sec
    sbc HI
    bne loo3
    bpl loo2
loo3:
    clc
    rts
loo2:
    sec
    lda X0
    sbc LO
    rts
cmpptr:
    sec
    lda up1 + 1
    sbc up2 + 1
    bne loo3
    bmi loo3
    lda up1
    sbc up2
    clc
    rts
cmplst:
    lda up1
    pha
    lda up1 + 1
    pha
    ldy #2
    lda (up1), y
    tax
    iny
    lda (up1), y
    sta up1 + 1
    stx up1
    lda up2
    pha
    lda up2 + 1
    pha
    lda (up2), y
    tax
    dey             ; 2
    lda (up2), y
    sta up2
    stx up2 + 1
    jsr order_rec
    bmi loo5
    beq loo6
loo5:
    php
    pla
    sta X0
    pla
    pla
    pla
    pla
    lda X0
    pha
    plp
    clc
    rts
loo6:
    pla
    sta X0 + 1
    pla 
    sta X0
    ldy #4
    lda (X0), y
    sta up2
    iny
    lda (X0), y
    sta up2 + 1
    pla
    sta X0 + 1
    pla
    sta X0
    lda (X0), y
    sta up1 + 1
    dey             ; 4
    lda (X0), y
    sta up1
    jmp order_rec
cmptup:
    ldy #1
    lda (up2), y
    sta X0
    lda (up1), y
    sec
    sbc X0
    beq cmplst
    clc
    rts
cmpstr:
    ldy #0
    lda (up1), y
    tax
    inx
strl1:
    lda (up1), y
    cmp (up2), y
    bne strl2
    iny
    dex
    bne strl1
strl2:
    clc
    rts

    ; E<I> -> Arg<J> -> A, <AP>=[I,J] (already multiplied by 2)
op_move_e_a:
    jsr fetch
    ldx X1 + 1
    ldy X1
    iny
    iny
    lda ENV, x
    sta LO
    sta (NGOAL), y
    inx
    iny
    lda ENV, x
    sta HI
    sta (NGOAL), y
    jmp op_addrefx

    ; mkvar -> E<I> -> Arg<J> -> A, <AP>=[I,J] (already multiplied by 2)
op_move_var_e_a:
    jsr fl_mkvar
    jsr fl_addref0
    jsr fetch
    ldx X1 + 1
    ldy X1
    iny
    iny
    lda LO
    sta ENV, x
    sta (NGOAL), y
    inx
    iny
    lda HI
    sta ENV, x
    sta (NGOAL), y
    jmp op_addrefx

    ; Arg<I> -> E<J> -> A, <AP>=[I,J] (already multiplied by 2)
op_move_a_e:
    jsr fetch
    ldy X1 + 1
    ldx X1
    iny
    iny
    lda (GOAL), y
    sta LO
    sta ENV, x
    iny
    inx
    lda (GOAL), y
    sta HI
    sta ENV, x
    jmp op_addrefx

    ; E<I> -> P<J> -> A, <AP>=[I,J] (already multiplied by 2)
op_move_e_p:
    jsr fetch
    ldx X1 + 1
    lda ENV, x
    sta LO
    inx
    lda ENV, x
    sta HI
    jsr fl_deref
    lda LO
    ldy X1
    sta PARAM0, y
    iny
    lda HI
    sta PARAM0, y
    jmp next

    ; mkvar -> E<I> -> P<J> -> A, <AP>=[I,J] (already multiplied by 2)
op_move_var_e_p:
    jsr fl_mkvar
    jsr fl_addref0
    jsr fetch
    ldx X1 + 1
    ldy X1
    lda LO
    sta ENV, x
    sta PARAM0, y
    inx
    iny
    lda HI
    sta ENV, x
    sta PARAM0, y
    jmp next

    ; A, <AP> = [Count, Val, Label, ...]
op_lookup:
    jsr fetch
    ldx X1
llok1:
    jsr fetch
    lda LO
    cmp X1
    bne llok2
    lda HI
    cmp X1 + 1
    bne llok2
    jmp op_jump
llok2:
    jsr skip
    dex
    bne llok1
    jmp op_jump

    ; x=ENV-index, A
fl_xassign:
    lda ENV, x
    sta X0
    inx
    lda ENV, x
    sta X0 + 1
    jmp fl_assign

    ; iterate over sstack
    ; (SP)=addr, invoke addr with entries in A(var), X0(saddr)
    ; (assumes sstack is not empty)
    ; modifies: X1, up1
for_sstack:
    pla
    clc
    adc #1
    sta X1
    pla
    adc #0
    sta X1 + 1
    lda #<sstack
    sta up1
    lda #>sstack
    sta up1 + 1
    lda sstackp     ; low byte
    lsr
    lsr
    tax
l40:
    ldy #0
    lda (up1), y
    sta LO
    iny
    lda (up1), y
    sta HI
    iny
    lda (up1), y
    sta X0
    iny
    lda (up1), y
    sta X0 + 1
    txa
    pha
    lda #>l41
    pha
    lda #<l41 - 1
    pha
    jmp (X1)
l41:
    clc
    lda up1
    adc #4
    sta up1
    pla
    tax
    dex
    bne l40
    rts

    ; write hex number
    ; A=num
writehex:
    lda HI
    jsr wbyte
    lda LO
wbyte:
    pha
    lsr
    lsr
    lsr
    lsr
    jsr wdigit
    pla
    and #$0f
wdigit:
    cmp #10
    bcc l110
    clc
    adc #"A" - "0" - 10
l110:
    clc
    adc #"0"
    jmp CHROUT

    ; convert integer to list
    ; P0=num, P1=list, P2=tail
fl_number_to_list_3:
    lda PARAM0
    sta LO
    lda PARAM0 + 1
    sta HI
    jsr fl_getint
    jsr _ItoA
    stx up1   
    ldx #6
consbuf:
    lda PARAM2
    sta X0
    lda PARAM2 + 1
    sta X0 + 1
ln2l3:
    lda #$40
    sta HI
    dex
    lda CnvStr, x
    asl
    rol HI
    ora #1
    sta LO
    lda #1 << LIST_BIT
    stx up2
    jsr fl_cons
    lda LO
    sta X0
    lda HI
    sta X0 + 1
    ldx up2
    cpx up1
    bne ln2l3
    lda PARAM1
    sta X0
    lda PARAM1 + 1
    sta X0 + 1
    jmp fl_unify_result

    ; A=int -> -A
    ; https://www.smspower.org/Development/Z80ProgrammingTechniques#x16bitNeg
negate:
    lda #0
    sec
    sbc LO
    sta ntmp
    sta LO
    sbc ntmp
    sec
    sbc HI
    sta HI
    rts
ntmp: .byt 0

    ; https://codebase64.org/doku.php?id=base:int16_and_uint16_conversion_to_string
bcd     = up1    ; system Fac, 3 bytes
sgn:     .byte 0
    .dsb 8 - (* & 7)
CnvStr:  .byt 0,0,0,0,0,0,0      ; 7 bytes: sgn + 5bytes + '\0'
 
;-------------------------------
; Converts a 16bit signed integer into string
;-------------------------------
; by Verz - Jul2019
;-------------------------------
;
;       Call with 16 bit number in LO/HI
;       result in CnvStr, X holds start index
;-------------------------------
 
_ItoA
        ldy #$0
        lda HI           ; is it positive?
        bpl _pos
        jsr negate
        ldy #'-'              ; prepares the sign value 
_pos    sty sgn
        jsr _UtoA             ; the work is done by _UtoA routine
        ldy sgn               ; if the sign was negative
        beq _enditoa          
        dex                   ; adds '-' before the number
        tya
        sta CnvStr,x
_enditoa
        rts
 
;-------------------------------
; Converts a 16bit unsigned integer into string
;-------------------------------
; by Verz - Jul2019
;-------------------------------
;
;       Call with 16 bit number in LO/HI
;       result in CnvStr
;-------------------------------
 
_UtoA
        jsr BINBCD16          ; converts the number to BCD
        ;lda bcd+2
        and #$0f              ; extracts every byte and adds $30
        ora #$30
        sta CnvStr+1
 
        lda bcd+1
        and #$0f
        ora #$30
        sta CnvStr+3
        lda bcd+1
        lsr
        lsr
        lsr
        lsr
        ora #$30
        sta CnvStr+2
 
        lda bcd+0
        and #$0f
        ora #$30
        sta CnvStr+5
        lda bcd+0
        lsr
        lsr
        lsr
        lsr
        ora #$30
        sta CnvStr+4
        lda #$20
        sta CnvStr+0
        ;rts                    ; decomment to avoid stripping leading 0s
 
        ldx #1                  ; remove 0s at beginning
_rem0   lda CnvStr,x
        cmp #$30                ; if it's a '0'
        bne _rts
        lda #$20                ; put a space instead
        sta CnvStr,x 
        inx
        cpx #$5                 ; exits before last digit
        bne _rem0
 
_rts    rts
 
;-------------------------------
; Converts a 16bit number to BCD
;-------------------------------
BINBCD16
        SED             ; Switch to decimal mode        2
        LDA #0          ; Ensure the result is clear    2
        STA bcd+0;                                      3
        STA bcd+1;                                      3
        STA bcd+2;                                      3       13
 
CBIT1   ASL LO       ; Shift out one bit             5
        ROL HI    ;                               5
;        LDA bcd+0      ;             
        ADC bcd+0       ; And add into result           3
        STA bcd+0       ;                               3
        ASL LO       ;                               5
        ROL HI     ;                               5
        ADC bcd+0       ;                               3
        STA bcd+0       ;                               3
        ASL LO       ;                               5
        ROL HI     ;                               5
        ADC bcd+0       ;                               3
        STA bcd+0       ;                               3
        ASL LO       ;                               5
        ROL HI     ;                               5
        ADC bcd+0       ;                               3
        STA bcd+0       ;                               3
        ASL LO       ;                               5
        ROL HI     ;                               5
        ADC bcd+0       ;                               3
        STA bcd+0       ;                               3
        ASL LO       ;                               5
        ROL HI     ;                               5
        ADC bcd+0       ;                               3
        STA bcd+0       ;                               3       96
 
        LDX #7;                                         2       2
CBIT7   ASL LO       ; Shift out one bit             5
        ROL HI     ;                               5
        LDA bcd+0       ; And add into result           3
        ADC bcd+0       ;                               3
        STA bcd+0       ;                               3
        LDA bcd+1       ; propagating any carry         3
        ADC bcd+1       ;                               3
        STA bcd+1       ;                               3
        DEX             ; And repeat for next bit       2
        BNE CBIT7       ;                               3       33*7-1=230
 
        LDX #3;                                         2       2
CBIT13  ASL LO       ; Shift out one bit             5
        ROL HI     ;                               5
        LDA bcd+0       ; And add into result           3
        ADC bcd+0       ;                               3
        STA bcd+0       ;                               3
        LDA bcd+1       ; propagating any carry         3
        ADC bcd+1       ;                               3
        STA bcd+1       ;                               3
        LDA bcd+2       ; ... thru whole result         3
        ADC bcd+2       ;                               3
        STA bcd+2       ;                               3
        DEX             ; And repeat for next bit       2
        BNE CBIT13      ;                               3       42*3-1=125
 
        CLD             ; Back to binary                2       2; tot 470
        rts             ; All Done.       

    ; list length
    ; A=list -> X0(lo)
    ; modifies: A
llength:
    lda #0
    sta X0
llen1:
    jsr fl_deref
    lda LO
    and #1 << IMMEDIATE_BIT
    beq llen2
llen3:
    rts
llen2:
    ldy #1
    lda (LO), y
    and #1 << LIST_BIT
    beq llen3
    inc X0
    ldy #4
    lda (LO), y
    tax
    iny
    lda (LO), y
    sta HI
    stx LO
    jmp llen1

fl_assign_3:
    lda PARAM0
    sta X0
    lda PARAM0 + 1
    sta X0 + 1
    lda PARAM1
    sta LO
    lda PARAM1
    sta HI
    jsr fl_assign
    lda PARAM2
    sta X0
    lda PARAM2 + 1
    sta X0 + 1
    ; fall through ...

    ; X0=var
confirm:
    lda #<fl_nil
    sta LO
    lda #>fl_nil
    sta HI
    jmp fl_assign

fl_list_to_tuple_2:
    lda PARAM0
    sta LO
    lda PARAM0 + 1
    sta HI
    jsr llength
    dec X0
    lda PARAM0
    sta LO
    lda PARAM0 + 1
    sta HI
    lda X0
    jsr fl_mktuple0
    lda PARAM1
    sta X0
    lda PARAM1 + 1
    sta X0 + 1
    jmp fl_unify_result

fl_halt_0:
    jmp exit

fl_uncons_3:
    lda PARAM0
    sta LO
    lda PARAM0 + 1
    sta HI
    jsr fl_check_cell
    ldy #2
    lda (PARAM0), y
    sta LO
    iny
    lda (PARAM0), y
    sta HI
    lda PARAM1
    sta X0
    lda PARAM1 + 1
    sta X0 + 1
    jsr fl_assign
    ldy #4
    lda (PARAM0), y
    sta LO
    iny
    lda (PARAM0), y
    sta HI
    lda PARAM2
    sta X0
    lda PARAM2 + 1
    sta X0 + 1
    jmp fl_assign

fl_string_to_list_3:
    lda PARAM0
    sta LO
    lda PARAM0 + 1
    sta HI
    jsr fl_check_atom
    lda PARAM2
    sta X0
    lda PARAM2 + 1
    sta X0 + 1
    ldy #0
    lda (PARAM0), y
    beq ls2l2
    tay
ls2l1:
    lda #$40
    sta HI
    lda (PARAM0), y
    asl
    ora #1
    sta LO
    rol HI
    lda #1 << LIST_BIT
    sty up1
    jsr fl_cons
    ldy up1
    lda LO
    sta X0
    lda HI
    sta X0 + 1
    dey
    bne ls2l1
ls2l2:
    lda PARAM1
    sta X0
    lda PARAM1 + 1
    sta X0 + 1
    jmp fl_unify_result

    ; locate and check index into tuple
    ; A=tup, X0=index -> A
index:
    dec X0
    beq lind2
    lda X0
    and #BITS_MASK >> 8
    beq lind1
    lda #<EBADINDEX
    sta X0
    lda #>EBADINDEX
    sta X0 + 1
    jmp error
lind1:
    ldy #4
    lda (LO), y
    tax
    iny
    lda (LO), y
    sta HI
    stx LO
    jsr fl_deref
    dec X0
    bne lind1
lind2:
    ldy #2
    lda (LO), y
    tax
    iny
    lda (LO), y
    sta HI
    stx LO
    rts

fl_get_arg_3:
    lda PARAM0
    sta LO
    lda PARAM0 + 1
    sta HI
    jsr fl_getint
    lda LO
    sta X0
    lda PARAM1
    sta LO
    lda PARAM1 + 1
    sta HI
    jsr fl_check_tuple
    jsr index
    lda PARAM2
    sta X0
    lda PARAM2 + 1
    sta X0 + 1
    jmp fl_unify_result

fl_put_arg_3:
    lda PARAM0
    sta LO
    lda PARAM0 + 1
    sta HI
    jsr fl_getint
    lda LO
    sta X0
    lda PARAM1
    sta LO
    lda PARAM1 + 1
    sta HI
    jsr fl_check_tuple
    jsr index
    lda LO
    sta X0
    lda HI
    sta X0 + 1
    lda PARAM2
    sta LO
    lda PARAM2 + 1
    sta HI
    jmp fl_assign

    ; chase stream tail and assign val 
    ; A=S, X1=val
    ; modifies: X0, A
chase_tail:
    lda LO
    sta X0
    lda HI
    sta X0 + 1
lct1:
    jsr fl_deref
    ldy #1
    lda (LO), y
    and #1 << LIST_BIT
    beq lct2    ; loop until non-list (expected: var) 
    ldy #4
    lda (LO), y
    tax
    iny
    lda (LO), y
    stx LO
    sta HI
    jmp lct1
lct2:
    lda LO
    pha
    lda HI
    pha
    lda X1
    sta LO
    lda X1 + 1
    sta HI
    jsr fl_addref   ; addref new tail
    jsr fl_unref    ; unref old S
    pla
    sta X0 + 1
    pla
    sta X0
    jmp fl_assign   ; assign X to tail

    ; P0=[S|_], P1=X - chase tail in S and assign X
fl_merge_2:
    ldy #2
    lda (PARAM0), y     ; get S
    sta LO
    iny
    lda (PARAM0), y
    sta HI
    lda PARAM1
    sta X1
    lda PARAM1 + 1
    sta X1 + 1
    jsr chase_tail
    lda PARAM1          ; destructively change S to X
    ldy #2
    sta (PARAM0), y
    lda PARAM1 + 1
    iny
    sta (PARAM0), y
    rts

    ; P0=[S|Count], P1=Amount
fl_merge_refs_2:
    ldy #5
    lda (PARAM0), y
    sec
    sbc #1
    clc
    adc PARAM1
    sta (PARAM0), y
    rts
    
    ; P0=addr, P1=byte
fl_poke_2:
    jsr getaddr
    lda PARAM1
    sta LO
    lda PARAM1 + 1
    sta HI
    jsr fl_check_integer
    jsr fl_getint
    lda LO
    ldy #0
    sta (X0), y
    rts

getaddr:
    lda PARAM0
    sta LO
    lda PARAM0 + 1
    sta HI
    jsr fl_check_integer
    jsr fl_getint
    lda LO
    sta X0
    lda HI
    sta X0 + 1
    rts

    ; P0=addr, P1=result
fl_peek_2:
    jsr getaddr
    ldy #0
    lda #$40
    sta HI
    lda (X0), y
    asl
    ora #1
    rol HI
    sta LO
    lda PARAM1
    sta X0
    lda PARAM1 + 1
    sta X0 + 1
    jmp fl_unify_result

; data area

ENOCELLS: .asc "HEAP FULL", 0
ENOGOALS: .asc "GOAL BUFFER FULL", 0
EQUEUEFULL: .asc "PROCESS POOL OVERFLOW", 0
EBOUND: .asc "ASSIGNMENT TO NON-VARIABLE", 0
EFAIL: .asc "FAILED", 0
EUNIFY: .asc "CAN NOT UNIFY", 0
EDIV0: .asc "DIVISION BY ZERO", 0
EDEADLOCK: .asc "DEADLOCK", 0
ELEAK: .asc "MEMORY LEAK", 0
ENOINT: .asc "EXPECTED INTEGER", 0
ENOSTRING: .asc "EXPECTED STRING", 0
ENOVAR: .asc "EXPECTED VARIABLE", 0
ENOTUPLE: .asc "EXPECTED TUPLE", 0
ENOLIST: .asc "EXPECTED LIST", 0
ENOCELL: .asc "EXPECTED CELL", 0
EBADINDEX: .asc "TUPLE INDEX OUT OF RANGE", 0
ESSFULL: .asc "SUSPENSION-STACK OVERFLOW", 0
ESTOPPED: .asc "STOPPED", 0
EARITY: .asc "TUPLE LENGTH MAXIMUM EXCEEDED', 0

ENV:    .dsb 2 * MAX_ENV

tmp1: .word 0
goals:  .word 0
cells:  .word 0
suspended: .word 0
ticks:  .word 0
stackbase: .word 0
idle:   .word 0
epoch:  .byte 2
tailcalls:  .byte 0
timeslice:  .byte TIMESLICE
freegoals: .word 0
freelist: .word 0

    .dsb * & 1
    .byte 0
fl_nil:
    .byte 2, "[]", 0
    .dsb * & 1
    .byte 0
fl_true:
    .byte 4, "true", 0
    .dsb * & 1
    .byte 0
fl_false:
    .byte 5, "false", 0

goalbuf = (HEAP - (MAX_GOALS + 1) * 5 * 2) & $ff00
sstack = goalbuf - 256
queuetop = sstack
queue = (queuetop - 2 * MAX_GOALS) & $ff00

init:
